from __future__ import annotations

import io
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional
import warnings

from gams.core.opt import (
    new_optHandle_tp,
    optCreateD,
    optReadDefinition,
    optClearMessages,
    optReadParameterFile,
    optFree,
    optGetMessage,
    GMS_SSSIZE,
    optMessageCount,
)
from pydantic import BaseModel, ConfigDict
from gamspy._config import get_option

from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy._model import Problem
    from types import FrameType

MULTI_SOLVE_MAP = {"replace": 0, "merge": 1, "clear": 2}
SOLVE_LINK_MAP = {"disk": 2, "memory": 5}

# GAMSPy to GAMS option mapping
OPTION_MAP = {
    "cns": "cns",
    "dnlp": "dnlp",
    "emp": "emp",
    "lp": "lp",
    "mcp": "mcp",
    "minlp": "minlp",
    "mip": "mip",
    "miqcp": "miqcp",
    "mpec": "mpec",
    "nlp": "nlp",
    "qcp": "qcp",
    "rminlp": "rminlp",
    "rmip": "rmip",
    "rmiqcp": "rmiqcp",
    "rmpec": "rmpec",
    "allow_suffix_in_equation": "suffixalgebravars",
    "allow_suffix_in_limited_variables": "suffixdlvars",
    "basis_detection_threshold": "bratio",
    "compile_error_limit": "cerr",
    "domain_violation_limit": "domlim",
    "hold_fixed_variables": "holdfixed",
    "iteration_limit": "iterlim",
    "keep_temporary_files": "keep",
    "listing_file": "output",
    "log_file": "logfile",
    "variable_listing_limit": "limcol",
    "equation_listing_limit": "limrow",
    "node_limit": "nodlim",
    "absolute_optimality_gap": "optca",
    "relative_optimality_gap": "optcr",
    "monitor_process_tree_memory": "procTreeMemMonitor",
    "memory_tick_interval": "procTreeMemTicks",
    "profile": "profile",
    "profile_file": "profileFile",
    "profile_tolerance": "profiletol",
    "reference_file": "reference",
    "time_limit": "reslim",
    "savepoint": "savepoint",
    "seed": "seed",
    "report_solution": "solprint",
    "show_os_memory": "showosmemory",
    "solve_link_type": "solvelink",
    "merge_strategy": "solveopt",
    "step_summary": "stepsum",
    "suppress_compiler_listing": "suppress",
    "report_solver_status": "sysout",
    "threads": "threads",
    "write_listing_file": "writeoutput",
    "zero_rounding_threshold": "zerores",
    "report_underflow": "zeroresrep",
}

MODEL_ATTR_OPTION_MAP = {
    "generate_name_dict": "dictfile",
    "enable_scaling": "scaleopt",
    "min_improvement_threshold": "cheat",
    "cutoff": "cutOff",
    "default_point": "defPoint",
    "enable_prior": "priorOpt",
    "infeasibility_tolerance": "tolInfRep",
    "try_partial_integer_solution": "tryInt",
    "examine_linearity": "tryLinear",
    "bypass_solver": "justscrdir",
}
EXECUTION_OPTIONS = {"loadpoint": "execute_loadpoint"}


class Options(BaseModel):
    """
    Options class to set GAMS options for the model.

    Attributes
    ----------

    cns: str | None
        Default **cns** solver
    dnlp: str | None
        Default **dnlp** solver
    emp: str | None
        Default **emp** solver
    lp: str | None
        Default **lp** solver
    mcp: str | None
        Default **mcp** solver
    minlp: str | None
        Default **minlp** solver
    mip: str | None
        Default **mip** solver
    miqcp: str | None
        Default **miqcp** solver
    mpec: str | None
        Default **mpec** solver
    nlp: str | None
        Default **nlp** solver
    qcp: str | None
        Default **qcp** solver
    rminlp: str | None
        Default **rminlp** solver
    rmip: str | None
        Default **rmip** solver
    rmiqcp: str | None
        Default **rmiqcp** solver
    rmpec: str | None
        Default **rmpec** solver
    allow_suffix_in_equation: bool | None
        Flag to allow variables with suffixes in model algebra
    allow_suffix_in_limited_variables: bool | None
        Flag to allow **domain limited variables** with suffixes in model
    basis_detection_threshold: float | None
        Basis detection threshold
    compile_error_limit: int = 1
        Compile time error limit
    domain_violation_limit: int | None
        Domain violation limit
    cutoff: float | None
        Within a branch-and-bound based solver, the parts of the tree with an objective value worse than
        the cutoff value are ignored. Note that this may speed up the initial phase of the branch and
        bound algorithm (before the first integer solution is found). However, the true optimum may be
        beyond the cutoff value. In this case the true optimum will be missed and moreover, no solution
        will be found. Observe that this option is specified in absolute terms.
    default_point: int | None
        This option determines the point that is passed to the solver as a basis. By default, the levels
        and marginals from the current basis are passed to the solver. In some circumstances (mostly during debugging),
        it can be useful to pass a standard default input point, i.e. with all levels set to 0 or lower bound.

        * **Option 0**	Pass user defined levels and marginals to solver

        * **Option 1**	Pass default levels and marginals to solver

        * **Option 2**	Pass default marginals to solver
    generate_name_dict: bool | None
        If this option is set, it will instruct GAMS to make the GAMS names
        of variables and equations that have been generated by the solve statement
        available to the solver. In many solver links, these names are registered
        with the solver and hence messages from the solver that involve variables
        and equations (e.g. an infeasible row or duplicate columns) can be easily
        interpreted by the user. However, the dictionary comes at a price.
        Generating the names and calculating and storing the map takes time and space.
        In addition, GAMS names take up space in the solver. Thus, if the user needs
        very fast generation and does not need names, setting dictFile to zero is a
        good option.
    enable_scaling: bool | None
        This option determines whether GAMS will employ user-specified variable and equation scaling factors.
        It must be set to True if scaling factors are to be used.
    enable_prior: bool | None
        Instructs the solver to use the priority branching information passed by GAMS through variable suffix
        values variable.prior. If and how priorities are used is solver-dependent.
    infeasibility_tolerance: float | None
        This option sets the tolerance for marking an equation infeasible in the equation listing. By default,
        1.0e-13.
    try_partial_integer_solution: bool | None
        Signals the solver to make use of a partial or near-integer-feasible solution stored in current variable
        values to get a quick integer-feasible point. The exact form of implementation depends on the solver and
        may be partly controlled by solver settings or options. See the solver manuals for details.
    examine_linearity: bool | None
        Examine empirical NLP model to see if there are any NLP terms active. If there are none the default LP
        solver will be used. If this option is set to True, empirical NLP models will be examined to determine
        if there are any active NLP terms. If there are none, the default LP solver will be used. The procedure
        also checks to see if QCP and DNLP models can be reduced to an LP; MIQCP and MINLP can be solved as a MIP;
        RMIQCP and RMINLP can be solved as an RMIP.
    bypass_solver: bool | None
        If True, GAMSPy does not pass the generated model to the solver. Useful for model generation time analysis.
    hold_fixed_variables: bool | None
        Treat fixed variables as constants
    iteration_limit: int | None
        Iteration limit of solver
    keep_temporary_files: bool = False
        Controls keeping or deletion of process directory and scratch files
    listing_file: str | None
        Listing file name
    loadpoint: str | None
        Path to the loadpoint GDX file that contains starting point records.
    log_file: str | None
        Log file name
    variable_listing_limit: int
        Maximum number of columns listed in one variable block
    equation_listing_limit: int
        Maximum number of rows listed in one equation block
    min_improvement_threshold: float | None
        For a branch-and-bound based solver, each new feasible solution must be at least the value of min_improvement_threshold
        better than the current best feasible solution. Note that this may speed up the search, but may cause some solutions,
        including optimal ones, to be missed. If a model has been solved with a nonzero min_improvement_threshold, then the
        optimal solution will be within the min_improvement_threshold or less of the found solution. Observe that the option
        min_improvement_threshold is specified in absolute terms, therefore non-negative values are appropriate for both
        minimization and maximization models.
    cutoff: float | None
        Within a branch-and-bound based solver, the parts of the tree with an objective value worse than the cutoff value are
        ignored. Note that this may speed up the initial phase of the branch and bound algorithm (before the first integer
        solution is found). However, the true optimum may be beyond the cutoff value. In this case the true optimum will be
        missed and moreover, no solution will be found.
    miro_protect:
        Protects MIRO input symbol records from being re-assigned, by default True.
    node_limit: int | None
        Node limit in branch and bound tree
    absolute_optimality_gap: float | None
        Absolute Optimality criterion solver default
    relative_optimality_gap: float | None
        Relative Optimality criterion solver default
    memory_tick_interval: float | None
        Wait interval between memory monitor checks: ticks = milliseconds
    monitor_process_tree_memory: bool | None
        Monitor the memory used by the GAMS process tree
    profile: int | None
        Execution profiling
    profile_file: str
        Write profile information to this file
    profile_tolerance: float | None
        Minimum time a statement must use to appear in profile generated output
    reference_file: str | None
        Symbol reference file
    time_limit: float | None
        Wall-clock time limit for solver
    savepoint: Optional[Literal[0, 1, 2, 3, 4]] = None
        Save solver point in GDX file

        * **Option 0**: No point GDX file is to be saved

        * **Option 1**: A point GDX file from the last solve is to be saved

        * **Option 2**: A point GDX file from every solve is to be saved

        * **Option 3**: A point GDX file from the last solve is to be saved in the scratch directory

        * **Option 4**: A point GDX file from every solve is to be saved in the scratch directory
    seed: int | None
        Random number seed
    report_solution: Literal[0, 1, 2] = 2
        Solution report print option

        * **Option 0**: Remove solution listings following solves

        * **Option 1**: Include solution listings following solves

        * **Option 2**: Suppress all solution information
    show_os_memory: Literal[0, 1, 2] = 0
        Show the memory usage reported by the Operating System instead of the internal counting

        * **Option 0**: Show memory reported by internal accounting

        * **Option 1**: Show resident set size reported by operating system

        * **Option 2**: Show virtual set size reported by operating system
    solve_link_type: Optional[Literal["disk", "memory"]] = None
        Solver link option

        * **disk**: Model instance saved to scratch directory, the solver is called with a spawn (if possible) or a shell (if spawn is not possible) while GAMS remains open.

        * **memory**: The model instance is passed to the solver in-memory - If this is not supported by the selected solver, it gets reset to **disk** automatically.
    merge_strategy: Optional[Literal["replace", "merge", "clear"]] = None
        * **Replace**: The solution information for all equations and variables is merged into the existing solution information

        * **Merge**: The solution information for all equations appearing in the model is completely replaced by the new model results; variables are only replaced if they appear in the final model

        * **Clear**: The solution information for all equations appearing in the model is completely replaced; in addition, variables appearing in the symbolic equations but removed by conditionals will be removed
    step_summary: bool | None
        Summary of computing resources used by job steps
    suppress_compiler_listing: bool = False
        Compiler listing option
    report_solver_status: bool | None
        Solver Status file reporting option
    threads: int | None
        Number of processors to be used by a solver
    write_listing_file: bool = True
        Switch to write a Listing file
    zero_rounding_threshold: float | None
        The results of certain operations will be set to zero if abs(result) LE ZeroRes
    report_underflow: bool | None
        Report underflow as a warning when abs(results) LE ZeroRes and result set to zero
    """

    model_config = ConfigDict(extra="forbid")

    cns: Optional[str] = None
    dnlp: Optional[str] = None
    emp: Optional[str] = None
    lp: Optional[str] = None
    mcp: Optional[str] = None
    minlp: Optional[str] = None
    mip: Optional[str] = None
    miqcp: Optional[str] = None
    mpec: Optional[str] = None
    nlp: Optional[str] = None
    qcp: Optional[str] = None
    rminlp: Optional[str] = None
    rmip: Optional[str] = None
    rmiqcp: Optional[str] = None
    rmpec: Optional[str] = None
    allow_suffix_in_equation: Optional[bool] = None
    allow_suffix_in_limited_variables: Optional[bool] = None
    basis_detection_threshold: Optional[float] = None
    compile_error_limit: Optional[int] = None
    domain_violation_limit: Optional[int] = None
    generate_name_dict: Optional[bool] = None
    enable_scaling: Optional[bool] = None
    enable_prior: Optional[bool] = None
    infeasibility_tolerance: Optional[float] = None
    try_partial_integer_solution: Optional[bool] = None
    examine_linearity: Optional[bool] = None
    bypass_solver: Optional[bool] = None
    min_improvement_threshold: Optional[float] = None
    cutoff: Optional[float] = None
    default_point: Optional[int] = None
    miro_protect: bool = True
    hold_fixed_variables: Optional[bool] = None
    iteration_limit: Optional[int] = None
    keep_temporary_files: Optional[int] = None
    listing_file: Optional[str] = None
    loadpoint: Optional[str] = None
    log_file: Optional[str] = None
    variable_listing_limit: int = 0
    equation_listing_limit: int = 0
    node_limit: Optional[int] = None
    absolute_optimality_gap: Optional[float] = None
    relative_optimality_gap: Optional[float] = None
    monitor_process_tree_memory: Optional[bool] = None
    memory_tick_interval: Optional[float] = None
    profile: Optional[int] = None
    profile_file: Optional[str] = None
    profile_tolerance: Optional[float] = None
    reference_file: Optional[str] = None
    time_limit: Optional[float] = None
    savepoint: Optional[Literal[0, 1, 2, 3, 4]] = None
    seed: Optional[int] = None
    report_solution: Literal[0, 1, 2] = 0
    show_os_memory: Optional[Literal[0, 1, 2]] = None
    solve_link_type: Optional[Literal["disk", "memory"]] = None
    merge_strategy: Optional[Literal["replace", "merge", "clear"]] = None
    step_summary: Optional[bool] = None
    suppress_compiler_listing: Optional[bool] = None
    report_solver_status: Optional[bool] = None
    threads: Optional[int] = None
    write_listing_file: Optional[bool] = None
    zero_rounding_threshold: Optional[float] = None
    report_underflow: Optional[bool] = None

    def model_post_init(self, context: Any) -> None:
        self._extra_options: dict[str, Any] = dict()
        self._debug_options: dict[str, Any] = dict()
        self._solver: str | None = None
        self._problem: str | None = None
        self._solver_options_file: str = "0"
        self._frame: FrameType | None = None

    def _get_gams_compatible_options(
        self, output: io.TextIOWrapper | None = None
    ) -> dict:
        gamspy_options = self.model_dump(exclude_none=True)
        if "allow_suffix_in_equation" in gamspy_options:
            allows_suffix = gamspy_options["allow_suffix_in_equation"]
            gamspy_options["allow_suffix_in_equation"] = (
                "on" if allows_suffix else "off"
            )

        if "allow_suffix_in_limited_variables" in gamspy_options:
            allows_suffix = gamspy_options["allow_suffix_in_limited_variables"]
            gamspy_options["allow_suffix_in_limited_variables"] = (
                "on" if allows_suffix else "off"
            )

        if "merge_strategy" in gamspy_options:
            strategy = gamspy_options["merge_strategy"]
            gamspy_options["merge_strategy"] = MULTI_SOLVE_MAP[strategy]

        if "solve_link_type" in gamspy_options:
            link_type = gamspy_options["solve_link_type"]
            gamspy_options["solve_link_type"] = SOLVE_LINK_MAP[link_type]

        if "listing_file" in gamspy_options:
            os.makedirs(
                Path(gamspy_options["listing_file"]).parent.absolute(),
                exist_ok=True,
            )

        if "log_file" in gamspy_options:
            os.makedirs(
                Path(gamspy_options["log_file"]).parent.absolute(),
                exist_ok=True,
            )

        gams_options = dict()
        for key, value in gamspy_options.items():
            if key not in OPTION_MAP:
                continue

            value = int(value) if isinstance(value, bool) else value
            gams_options[OPTION_MAP[key]] = value

        gams_options["previouswork"] = (
            1  # # In case GAMS version differs on backend
        )
        gams_options["traceopt"] = 3

        if self.log_file:
            if output is not None:
                gams_options["logoption"] = 4
            else:
                gams_options["logoption"] = 2
        else:
            if output is not None:
                gams_options["logoption"] = 3
            else:
                gams_options["logoption"] = 0

        return gams_options

    def _set_solver_options(
        self,
        system_directory: str,
        working_directory: str,
        solver: str,
        problem: Problem,
        solver_options: dict | None,
    ):
        """Set the solver and the solver options"""
        self._solver = solver
        self._problem = str(problem)

        if solver_options:
            write_solver_options(
                system_directory, working_directory, solver, solver_options
            )
            self._solver_options_file = "1"
        else:
            self._solver_options_file = "0"

    def _set_extra_options(self, options: dict) -> None:
        """Set extra options of the backend"""
        self._extra_options = options

    def _set_debug_options(self, options: dict) -> None:
        """Set debugging options"""
        self._debug_options = options

    @staticmethod
    def fromFile(path: str) -> Options:
        """
        Generates an Options object with the key-value pairs in a file.
        The file in given path must consist of one key-value pair in each line.

        Parameters
        ----------
        path : str
            Path to the option file.

        Returns
        -------
        Options

        Raises
        ------
        ValidationError
            In case the given path is not a file.
        """
        if not os.path.isfile(path):
            raise ValidationError(f"No such file in the given path: {path}")

        attributes = dict()
        with open(path, encoding="utf-8") as file:
            lines = file.readlines()

            for line in lines:
                if line == "\n" or line == "":
                    continue  # pragma: no cover

                key, value = line.split("=")
                attributes[key.strip()] = value.strip()

        return Options(**attributes)

    def export(self, pf_file: str) -> None:
        """
        Exports options to the pf_file. Each line contains a key-value pair.

        Parameters
        ----------
        pf_file : str
        """
        self._export(pf_file)

    def _export(
        self, pf_file: str, output: io.TextIOWrapper | None = None
    ) -> None:
        """
        Exports options to the pf_file. Each line contains a key-value pair.

        Parameters
        ----------
        pf_file : str
        output : io.TextIOWrapper | None, optional
        """
        all_options = dict()
        # Solver options
        if self._solver is not None:
            all_options[self._problem] = self._solver

        all_options["optfile"] = self._solver_options_file

        # Extra options
        all_options.update(**self._extra_options)
        all_options.update(**self._debug_options)

        if self._frame is not None:
            filename = self._frame.f_code.co_filename
            line_number = self._frame.f_lineno
            all_options["GP_SolveLine"] = f"{filename} line {line_number}"

        # User options
        user_options = self._get_gams_compatible_options(output)
        all_options.update(**user_options)

        # Generate pf file
        with open(pf_file, "w", encoding="utf-8") as file:
            file.write(
                "\n".join(
                    [
                        f'{key} = "{value}"'
                        for key, value in all_options.items()
                    ]
                )
            )

class FreezeOptions(BaseModel):
    no_match_limit: int = 0
    debug: bool = False
    update_type: Literal["0", "base_case", "accumulate", "inherit"] = (
        "base_case"
    )

class ModelInstanceOptions(BaseModel):
    no_match_limit: int = 0
    debug: bool = False
    update_type: Literal["0", "base_case", "accumulate", "inherit"] = (
        "base_case"
    )

    def __init__(self, **kwargs):
        warnings.warn(
            "ModelInstanceOptions will be renamed to FreezeOptions in GAMSPy 1.9.0. Please use FreezeOptions instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(**kwargs)

class ConvertOptions(BaseModel):
    """
    Options for the conversion of GAMSPy models into different formats.

    Attributes
    ----------
    AmplNLBin : Optional[bool]
        Enables binary .nl file. False by default.
    AmplNlInitDual : Optional[int]
        Specifies which initial equation marginal values to write to the .nl file.
        
        * **0**: Write no values 
        
        * **1**: Write only nondefault values
        
        * **2**: Write all values.

        1 by default.
    AmplNlInitPrimal : Optional[int]
        Specifies which initial primal values to write to the .nl file.
        
        * **0**: Write no values
        
        * **1**: Write only nondefault values
        
        * **2**: Write all values.

        2 by default.
    GDXHessian : Optional[bool]
        Controls whether Hessian information is included in GDX Jacobian file.
        False by default.
    GDXNames : Optional[bool]
        Controls whether variable and equation names are included in GDX Jacobian file.
        True by default.
    GDXQuadratic : Optional[bool]
        Specifies whether quadratic terms are included in GDX Jacobian file.
        False by default.
    GDXUELs : Optional[bool]
        Controls whether Universal Element List (UEL) information is included in GDX Jacobian file.
        True by default.
    GAMSInsert : Optional[str]
        Allows the insertion of custom GAMS code into the model.
    HeaderTimeStamp : Optional[bool]
        Specifies a timestamp to include in the header of the output file.
    GDXIntervalEval : Optional[bool]
        Controls the inclusion of interval evaluation (symbols `A_int` and `e_int`) into the GDX Jacobian format.
        False by default.
    GAMSObjVar : Optional[str]
        Specifies the name of the objective variable in the GAMS scalar model.
    PermuteEqus : Optional[bool]
        Enables or disables the permutation of equations.
        False by default.
    PermuteVars : Optional[bool]
        Enables or disables the permutation of variables.
        False by default.
    QExtractAlg : Optional[int]
        Specifies the algorithm used for quadratic extraction.
        
        * **0**: Automatic

        * **1**: ThreePass: Uses a three-pass forward / backward / forward AD technique to compute function / gradient / Hessian values and a hybrid scheme for storage.

        * **2**: DoubleForward: Uses forward-mode AD to compute and store function, gradient, and Hessian values at each node or stack level as required. The gradients and Hessians are stored in linked lists.

        * **3**: Concurrent: Uses ThreePass and DoubleForward in parallel. As soon as one finishes, the other one stops.

        0 by default.
    Reform : Optional[int]
        Controls the reformulation of certain structures in the model.
        0: No reformulation, 1: Apply reformulation.
    SkipNRows : Optional[int]
        Skip constraints of type `NONBINDING`.
    Width : Optional[int]
        Sets the width for certain output formats, such as tables or reports.
    """
    
    model_config = ConfigDict(extra="forbid")

    AmplNLBin: Optional[bool] = None
    AmplNlInitDual: Optional[int] = None
    AmplNlInitPrimal: Optional[int] = None
    GDXHessian: Optional[bool] = None
    GDXNames: Optional[bool] = None
    GDXQuadratic: Optional[bool] = None
    GDXUELs: Optional[bool] = None
    GAMSInsert: Optional[str] = None
    HeaderTimeStamp: Optional[bool] = None
    GDXIntervalEval: Optional[bool] = None
    GAMSObjVar: Optional[str] = None
    PermuteEqus: Optional[bool] = None
    PermuteVars: Optional[bool] = None
    QExtractAlg: Optional[int] = None
    Reform: Optional[int] = None
    SkipNRows: Optional[int] = None
    Width: Optional[int] = None


def write_solver_options(
    system_directory: str,
    working_directory: str,
    solver: str,
    solver_options: dict,
) -> None:
    options_file_name = os.path.join(
        working_directory, f"{solver.lower()}.opt"
    )
    with open(options_file_name, "w", encoding="utf-8") as solver_file:
        for key, value in solver_options.items():
            row = f"{key} {value}\n"
            if solver.upper() in ("SHOT", "SOPLEX", "SCIP", "HIGHS"):
                row = f"{key} = {value}\n"

            solver_file.write(row)

    # The following solvers do not use the opt<solver>.def file
    if solver.upper() in (
        "HIGHS",
        "SOPLEX",
        "KESTREL",
        "SCIP",
        "SHOT",
        "SOPLEX",
    ):
        return

    if get_option("VALIDATION") and get_option("SOLVER_OPTION_VALIDATION"):
        validate_solver_options(system_directory, options_file_name, solver)


def get_def_file(system_directory: str, solver: str) -> str:
    solver_name = solver.upper()
    def_file_path = os.path.join(system_directory, f"opt{solver.lower()}.def")

    if solver_name == "CONOPT4":
        return os.path.join(system_directory, "optconopt.def")

    if solver_name == "EXAMINER2":
        return os.path.join(system_directory, "optexaminer.def")
    
    if solver_name == "IPOPTH":
        return os.path.join(system_directory, "optipopt.def")

    return def_file_path


def validate_solver_options(
    system_directory: str, options_file_name: str, solver: str
) -> None:
    if not get_option("VALIDATION"):
        return

    option_handle = new_optHandle_tp()
    rc, msg = optCreateD(option_handle, system_directory, GMS_SSSIZE)

    # Return code 0 means there is an error. Weird but this is what we have to work with.
    if rc == 0:
        raise RuntimeError(msg)

    # Check the validity of the .def file.
    solver_def_file = get_def_file(system_directory, solver)
    if optReadDefinition(option_handle, solver_def_file):
        msg_list = []
        for i in range(optMessageCount(option_handle)):
            msg_list.append(optGetMessage(option_handle, i + 1))

        raise RuntimeError(
            f"Error while processing {solver_def_file}. Log messages: {msg_list}"
        )

    optClearMessages(option_handle)

    # Check the validity of the parameters 
    if optReadParameterFile(option_handle, options_file_name):
        raise RuntimeError(f"Error while reading {options_file_name}")

    msg_list = []
    for i in range(optMessageCount(option_handle)):
        msg_list.append(optGetMessage(option_handle, i + 1))

    optClearMessages(option_handle)
    optFree(option_handle)

    if msg_list:
        error_messages = []
        for message in msg_list:
            # optMsgInputEcho    = 0,
            # optMsgHelp         = 1,
            # optMsgDefineError  = 2,
            # optMsgValueError   = 3,
            # optMsgValueWarning = 4,
            # optMsgDeprecated   = 5,
            # optMsgFileEnter    = 6,
            # optMsgFileLeave    = 7,
            # optMsgTooManyMsgs  = 8,
            # optMsgUserError    = 9
            if message[1] not in (6, 7):
                error_messages.append(message[0])

        if error_messages:
            error_message = "\n".join(error_messages)
            raise ValidationError(
                f"Error while reading the parameter file: \n\n{error_message}"
            )



from __future__ import annotations

import base64
import logging
import os
import shutil
import ssl
import xmlrpc.client
import zipfile
from typing import TYPE_CHECKING

import certifi

import gamspy._backend.backend as backend
import gamspy.utils as utils
from gamspy._options import Options
from gamspy.exceptions import (
    GamspyException,
    NeosClientException,
    ValidationError,
)

logger = logging.getLogger("NEOS")
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)
formatter = logging.Formatter("[%(name)s - %(levelname)s] %(message)s")
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)

if TYPE_CHECKING:
    import io

    from gamspy import Container, Model
    from gamspy._symbols.symbol import Symbol


class NeosClient:
    def __init__(
        self,
        email: str,
        server: str = "https://neos-server.org:3333",
        username: str | None = None,
        password: str | None = None,
        priority: str = "long",
        is_blocking: bool = True,
    ) -> None:
        """
        Python client for NEOS Server. Implements the functions listed in:
        https://neos-server.org/neos/xml-rpc.html

        Parameters
        ----------
        email : str
        server : _type_, optional
            Server ip and port, by default "https://neos-server.org:3333"
        username : Optional[str], optional
            Username, by default None
        password : Optional[str], optional
            Password, by default None
        priority : str, optional
            Priority of the job, by default "long"
        is_blocking : bool, optional
            Decides on sync or async execution, by default True
        """
        self.email = email
        self.server = server
        self.username = username
        self.password = password
        self.priority = priority
        self.is_blocking = is_blocking
        context = ssl.create_default_context(cafile=certifi.where())
        self.neos = xmlrpc.client.ServerProxy(server, context=context)
        self.jobs: list[tuple] = []

    def is_alive(self) -> bool:
        """
        Checks if NEOS Server is alive

        Returns
        -------
        bool
        """
        response: str = self.neos.ping()
        return bool(response.startswith("NeosServer is alive"))

    def get_job_status(self, job_number: int, job_password: str) -> str:
        """
        Returns the status of the job.

        Parameters
        ----------
        job_number : int
        job_password : str

        Returns
        -------
        str
            Either "Done", "Running", "Waiting", "Unknown Job", or "Bad Password"
        """
        return self.neos.getJobStatus(job_number, job_password)

    def get_completion_code(self, job_number: int, job_password: str) -> str:
        """
        Gets the completion code for "Done" jobs. Result is undefined for jobs
        that are "Waiting" or "Running". Returns "Normal", "Out of memory", "Timed out",
        "Disk Space", "Server error", "Unknown Job", or "Bad Password"

        Parameters
        ----------
        job_number : int
        job_password : str

        Returns
        -------
        str
        """
        return self.neos.getCompletionCode(job_number, job_password)

    def get_job_info(self, job_number: int, job_password: str) -> tuple:
        """
        Gets information about the job.

        Parameters
        ----------
        job_number : int
        job_password : str

        Returns
        -------
        tuple
            (category, solver_name, input, status, completion_code)
        """
        return self.neos.getJobInfo(job_number, job_password)

    def kill_job(self, job_number: int, job_password: str, killmsg="") -> str:
        """
        Cancel a submitted job that is running or waiting to run on NEOS.

        Parameters
        ----------
        job_number : int
        job_password : str
        killmsg : str, optional

        Returns
        -------
        str
        """
        return self.neos.killJob(job_number, job_password, killmsg)

    def get_final_results(
        self, job_number: int, job_password: str, is_blocking: bool = True
    ) -> xmlrpc.client.Binary:
        """
        Retrieve results from a submitted job on NEOS. If the job is still
        running and the user is not authenticated, then this function will
        hang until the job is finished. The function returns a base-64 encoded object.

        Parameters
        ----------
        job_number : int
        job_password : str
        is_blocking : bool, optional

        Returns
        -------
        xmlrpc.client.Binary
        """
        if is_blocking:
            return self.neos.getFinalResults(job_number, job_password)

        return self.neos.getFinalResultsNonBlocking(job_number, job_password)

    def email_job_results(self, job_number: int, job_password: str) -> str:
        """
        Results for a finished job will be emailed to the email address specified in the job submission.
        If results are too large for email, they will not be emailed (though they can be accessed via
        the NEOS website)


        Parameters
        ----------
        job_number : int
        job_password : str

        Returns
        -------
        str
        """
        return self.neos.emailJobResults(job_number, job_password)

    def get_intermediate_results(
        self,
        job_number: int,
        job_password: str,
        offset: int,
        is_blocking: bool = True,
    ) -> xmlrpc.client.Binary:
        """
        Gets intermediate results of a job submitted to NEOS, starting at the specified character offset up to
        the last received data. Intermediate results are usually the standard output of the solver daemon.
        Note that because output does not stream for jobs with "long" priority (default value), getIntermediateResults()
        will not return any results for long priority jobs. Output does stream for jobs with "short" priority
        (maximum time of 5 minutes).

        Parameters
        ----------
        job_number : int
        job_password : str
        offset : int
        is_blocking : bool, optional

        Returns
        -------
        xmlrpc.client.Binary
        """
        if is_blocking:
            return self.neos.getIntermediateResults(
                job_number, job_password, offset
            )

        return self.neos.getIntermediateResultsNonBlocking(
            job_number, job_password
        )

    def download_output(
        self,
        job_number: int,
        job_password: str,
        working_directory: str = ".",
    ) -> None:
        """
        Downloads the output of the job and writes it in a zip file.

        Parameters
        ----------
        job_number : int
        job_password : str
        working_directory: str
        """
        filename = f"{job_number}-{job_password}-solver-output.zip"
        try:
            os.makedirs(working_directory)
        except FileExistsError:
            pass

        response = self.neos.getOutputFile(job_number, job_password, filename)
        if str(response) == "Output file does not exist":
            raise NeosClientException(
                "Couldn't get output file from NEOS Server because:"
                f" {response}"
            )

        with open(os.path.join(working_directory, filename), "wb") as file:
            file.write(response.data)

        with zipfile.ZipFile(
            os.path.join(working_directory, filename), "r"
        ) as zip_ref:
            zip_ref.extractall(working_directory)

    def _prepare_xml(
        self,
        gams_string: str,
        solver_options: dict | None,
        gdx_path: str,
        restart_path: str,
        options: Options,
        save_name: str | None = None,
        xml_path: str = "neos.xml",
        working_directory: str = ".",
    ) -> None:
        gdx_string = ""
        if os.path.exists(gdx_path):
            with open(gdx_path, "rb") as gdx_file:
                content = gdx_file.read()
                gdx_base64 = base64.b64encode(content).decode("utf-8")
                gdx_string = f"<base64>{gdx_base64}</base64>"

        restart_string = ""
        try:
            with open(restart_path, "rb") as restart_file:
                content = restart_file.read()
                restart_base64 = base64.b64encode(content).decode("utf-8")
                restart_string = f"<base64>{restart_base64}\n</base64>"
        except FileNotFoundError:
            pass

        solver_options_str = ""
        if solver_options:
            solver_options_str = "\n".join(
                [f"{key} {value}" for key, value in solver_options.items()]
            )

        options._export(os.path.join(working_directory, "parameters"))
        with open(
            os.path.join(working_directory, "parameters"), encoding="utf-8"
        ) as file:
            parameters = [line.rstrip() for line in file.readlines()]

        extras = []
        if save_name is not None:
            extras.append(f"save={save_name}")

        parameter_string = "\n".join(parameters + extras)

        template = f"""
            <document>
            <category>lp</category>
            <solver>cplex</solver>
            <inputType>GAMS</inputType>
            <email>{self.email}</email>
            <priority>{self.priority}</priority>
            <model><![CDATA[{gams_string}]]></model>
            <options><![CDATA[{solver_options_str}]]></options>
            <parameters><![CDATA[{parameter_string}]]></parameters>
            <gdx>{gdx_string}</gdx>
            <restart>{restart_string}</restart>
            <wantgdx><![CDATA[]]></wantgdx>
            <wantlst><![CDATA[yes]]></wantlst>
            <wantlog><![CDATA[yes]]></wantlog>
            </document>
        """

        with open(
            os.path.join(working_directory, xml_path), "w", encoding="utf-8"
        ) as neos_xml:
            neos_xml.write(template)

    def print_queue(self):
        """Prints NEOS Server queue"""
        if not self.is_alive():
            raise NeosClientException(
                "NeosServer is not alive. Try again later."
            )

        msg = self.neos.printQueue()
        print(msg)

    def submit_job(
        self,
        xml_path: str = "neos.xml",
        working_directory: str = ".",
    ) -> tuple[int, str]:
        """
        Submits the job to NEOS Server.

        Parameters
        ----------
        xml_path : str, optional
        is_blocking : bool, optional

        Returns
        -------
        Tuple[int, str]
            Job number and job password

        Raises
        ------
        NeosClientException
            In case there was an error on NeosServer
        """
        with open(
            os.path.join(working_directory, xml_path), encoding="utf-8"
        ) as file:
            xml = file.read()

        if not self.is_alive():
            raise NeosClientException(
                "NeosServer is not alive. Try again later."
            )

        if self.username is not None and self.password is not None:
            job_number, job_password = self.neos.authenticatedSubmitJob(
                xml, self.username, self.password
            )
        else:
            job_number, job_password = self.neos.submitJob(xml)

        logger.info(f"Job Number: {job_number}, Job Password: {job_password}")
        self.jobs.append((job_number, job_password))

        if job_number == 0:
            raise NeosClientException(f"NEOS Server error! {job_password}")

        return job_number, job_password


class NEOSServer(backend.Backend):
    def __init__(
        self,
        container: Container,
        options: Options,
        solver: str,
        solver_options: dict | None,
        client: NeosClient | None,
        output: io.TextIOWrapper | None,
        model: Model,
        load_symbols: list[Symbol] | None,
    ) -> None:
        if client is None or not isinstance(client, NeosClient):
            raise ValidationError(
                "`NeosClient` must be provided to solve on NEOS Server."
            )

        super().__init__(
            container,
            model,
            options,
            solver,
            solver_options,
            output,
            load_symbols,
        )
        self.client = client
        self.job_name = self.get_job_name()
        self.gms_file = self.job_name + ".gms"
        self.pf_file = self.job_name + ".pf"
        self.restart_file = self.job_name + ".g00"
        self.trace_file = self.job_name + ".txt"

    def is_async(self):
        return not self.client.is_blocking

    def run(
        self,
        relaxed_domain_mapping: bool = False,
        gams_to_gamspy: bool = False,
    ):
        # Run a dummy job to get the restart file to be sent to NEOS Server
        self._create_restart_file()

        # Generate gams string and write modified symbols to gdx
        gams_string = self.preprocess("in.gdx")

        # Run the model
        self.execute_gams(gams_string)

        if self.is_async():
            return None

        # Synchronize GAMSPy with checkpoint and return a summary
        summary = self.postprocess()

        # Run another dummy job to synchronize GAMS and GAMSPy state
        self._sync()

        return summary

    def execute_gams(self, gams_string: str):
        working_directory = self.container.working_directory
        if self.container._debugging_level == "keep":
            self.options.log_file = os.path.basename(self.job_name) + ".log"

        extra_options = {
            "gdx": "output.gdx",
            "gdxSymbols": "newOrChanged",
            "trace": os.path.basename(self.trace_file),
            "forcework": "1",
        }
        self.options._set_extra_options(extra_options)

        self.client._prepare_xml(
            gams_string,
            solver_options=self.solver_options,
            gdx_path=self.container._gdx_in,
            restart_path=self.restart_file,
            options=self.options,
            working_directory=working_directory,
        )

        job_number, job_password = self.client.submit_job(
            working_directory=working_directory
        )

        if self.client.is_blocking:
            _ = self.client.get_final_results(job_number, job_password)

            self.client.download_output(
                job_number,
                job_password,
                working_directory=working_directory,
            )

            out_gdx = os.path.join(working_directory, "output.gdx")
            log_path = os.path.join(working_directory, "solve.log")
            if not os.path.exists(out_gdx):
                raise GamspyException(
                    "The job was not completed successfully. Check"
                    f" {log_path} for details."
                )
            shutil.move(out_gdx, self.container._gdx_out)

            if self.output is not None:
                with open(log_path) as file:
                    self.output.write(file.read())

            if self.options.log_file is not None and not os.path.exists(
                self.options.log_file
            ):
                shutil.copy(log_path, self.options.log_file)

        self.container._unsaved_statements = []
        if not self.is_async() and self.model:
            self.model._update_model_attributes()
            self.container._delete_autogenerated_symbols()

    def postprocess(self):
        super().load_records()
        if self.client.is_blocking:
            self.parse_listings()
            return self.prepare_summary(self.trace_file)

        return None

    def _prepare_dummy_options(self) -> dict:
        scrdir = self.container._process_directory

        extra_options = {
            "gdx": self.container._gdx_out,
            "gdxSymbols": "newOrChanged",
            "trace": self.trace_file,
            "input": self.gms_file,
            "sysdir": self.container.system_directory,
            "scrdir": scrdir,
            "scriptnext": os.path.join(scrdir, "gamsnext.sh"),
            "previouswork": 1,
            "license": utils._get_license_path(
                self.container.system_directory
            ),
        }

        if self.container._network_license:
            extra_options["netlicense"] = os.path.join(scrdir, "gamslice.dat")

        return extra_options

    def _create_restart_file(self):
        with open(self.gms_file, "w", encoding="utf-8") as gams_file:
            gams_file.write("")

        options = Options()
        extra_options = self._prepare_dummy_options()
        options._set_extra_options(extra_options)
        options._extra_options["save"] = self.restart_file
        options._export(self.pf_file)

        self.container._send_job(self.job_name, self.pf_file)

    def _sync(self):
        symbols = utils._get_symbol_names_from_gdx(
            self.container.system_directory, self.container._gdx_out
        )
        dirty_str = ",".join(symbols)
        with open(self.gms_file, "w", encoding="utf-8") as gams_file:
            gams_file.write(
                f'execute_load "{self.container._gdx_out}", {dirty_str};'
            )

        options = Options()
        extra_options = self._prepare_dummy_options()
        options._set_extra_options(extra_options)
        options._export(self.pf_file)

        self.container._send_job(self.job_name, self.pf_file)



from __future__ import annotations

from gamspy.math.activation import (
    log_softmax,
    relu_with_binary_var,
    relu_with_complementarity_var,
    relu_with_sos1_var,
    softmax,
)
from gamspy.math.log_power import (
    cv_power,
    exp,
    log,
    log2,
    log10,
    log_beta,
    log_gamma,
    logit,
    power,
    rpower,
    sign_power,
    sllog10,
    sqlog10,
    sqr,
    vc_power,
)
from gamspy.math.matrix import (
    Dim,
    _generate_dims,
    dim,
    next_alias,
    permute,
    trace,
    vector_norm,
)
from gamspy.math.misc import (
    MathOp,
    Max,
    Min,
    Round,
    abs,
    beta,
    bool_and,
    bool_eqv,
    bool_imp,
    bool_not,
    bool_or,
    bool_xor,
    ceil,
    dist,
    div,
    div0,
    entropy,
    errorf,
    factorial,
    floor,
    fractional,
    gamma,
    ifthen,
    lse_max,
    lse_max_sc,
    lse_min,
    lse_min_sc,
    map_value,
    mod,
    ncp_cm,
    ncp_f,
    ncpVUpow,
    ncpVUsin,
    poly,
    rand_binomial,
    rand_linear,
    rand_triangle,
    regularized_beta,
    regularized_gamma,
    rel_eq,
    rel_ge,
    rel_gt,
    rel_le,
    rel_lt,
    rel_ne,
    same_as,
    sigmoid,
    sign,
    slexp,
    slrec,
    sqexp,
    sqrec,
    sqrt,
    truncate,
)
from gamspy.math.probability import (
    binomial,
    centropy,
    normal,
    uniform,
    uniformInt,
)
from gamspy.math.trigonometric import (
    acos,
    asin,
    atan,
    atan2,
    cos,
    cosh,
    sin,
    sinh,
    tan,
    tanh,
)

__all__ = [
    "cos",
    "cosh",
    "sin",
    "sinh",
    "acos",
    "asin",
    "tan",
    "tanh",
    "atan",
    "atan2",
    "exp",
    "power",
    "sqr",
    "sqrt",
    "log",
    "log2",
    "log10",
    "log_beta",
    "log_gamma",
    "logit",
    "permute",
    "trace",
    "vector_norm",
    "dim",
    "Dim",
    "next_alias",
    "abs",
    "ceil",
    "dist",
    "div",
    "div0",
    "factorial",
    "floor",
    "fractional",
    "Min",
    "MathOp",
    "Max",
    "mod",
    "Round",
    "sign",
    "binomial",
    "centropy",
    "normal",
    "uniform",
    "uniformInt",
    "cv_power",
    "rpower",
    "sign_power",
    "sllog10",
    "sqlog10",
    "vc_power",
    "slexp",
    "sqexp",
    "truncate",
    "ifthen",
    "beta",
    "regularized_beta",
    "gamma",
    "regularized_gamma",
    "entropy",
    "lse_max",
    "lse_max_sc",
    "lse_min",
    "lse_min_sc",
    "ncp_cm",
    "ncp_f",
    "ncpVUpow",
    "ncpVUsin",
    "poly",
    "rand_binomial",
    "rand_linear",
    "rand_triangle",
    "slrec",
    "sqrec",
    "errorf",
    "same_as",
    "sigmoid",
    "bool_and",
    "bool_eqv",
    "bool_imp",
    "bool_not",
    "bool_or",
    "bool_xor",
    "rel_eq",
    "rel_ge",
    "rel_gt",
    "rel_le",
    "rel_lt",
    "rel_ne",
    "_generate_dims",
    "relu_with_sos1_var",
    "relu_with_binary_var",
    "relu_with_complementarity_var",
    "log_softmax",
    "softmax",
    "map_value",
]



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import gamspy._backend.backend as backend
import gamspy._miro as miro
import gamspy.utils as utils
from gamspy.exceptions import GamspyException, _customize_exception

if TYPE_CHECKING:
    import io

    from gamspy import Container, Model, Options
    from gamspy._symbols.symbol import Symbol


class Local(backend.Backend):
    def __init__(
        self,
        container: Container,
        options: Options,
        solver: str | None,
        solver_options: dict | None,
        output: io.TextIOWrapper | None,
        model: Model | None,
        load_symbols: list[Symbol] | None,
    ) -> None:
        super().__init__(
            container,
            model,
            options,
            solver,
            solver_options,
            output,
            load_symbols,
        )
        self.job_name = self.get_job_name()
        self.gms_file = self.job_name + ".gms"
        self.lst_file = self.job_name + ".lst"
        self.pf_file = self.job_name + ".pf"
        self.trace_file = self.job_name + ".txt"

    def _prepare_extra_options(self, gams_to_gamspy: bool) -> dict:
        scrdir = self.container._process_directory

        extra_options = {
            "trace": self.trace_file,
            "input": self.gms_file,
            "output": self.lst_file,
            "optdir": self.container.working_directory,
            "sysdir": self.container.system_directory,
            "scrdir": scrdir,
            "scriptnext": os.path.join(scrdir, "gamsnext.sh"),
            "license": utils._get_license_path(
                self.container.system_directory
            ),
        }

        if gams_to_gamspy:
            extra_options["gdx"] = self.container._gdx_out
            extra_options["gdxSymbols"] = "newOrChanged"

        if self.container._network_license:
            extra_options["netlicense"] = os.path.join(scrdir, "gamslice.dat")

        return extra_options

    def is_async(self):
        return False

    def run(
        self,
        relaxed_domain_mapping: bool = False,
        gams_to_gamspy: bool = False,
    ):
        # Generate gams string and write modified symbols to gdx
        gams_string = self.preprocess(self.container._gdx_in)

        # Run the model
        self.execute_gams(gams_string, gams_to_gamspy)

        # Synchronize GAMSPy with checkpoint and return a summary
        summary = self.postprocess(relaxed_domain_mapping, gams_to_gamspy)

        return summary

    def execute_gams(self, gams_string: str, gams_to_gamspy: bool):
        # Write gms file
        with open(self.gms_file, "w", encoding="utf-8") as gams_file:
            gams_file.write(gams_string)

        # Write pf file
        extra_options = self._prepare_extra_options(gams_to_gamspy)
        self.options._set_extra_options(extra_options)
        self.options._export(self.pf_file, self.output)

        try:
            self.container._job = self.job_name
            self.container._send_job(self.job_name, self.pf_file, self.output)

            if not self.is_async() and self.model:
                self.model._update_model_attributes()
                self.container._delete_autogenerated_symbols()
        except GamspyException as exception:
            self.container._workspace._errors.append(str(exception))
            message = _customize_exception(
                self.options,
                self.job_name,
                exception.return_code,
            )

            exception.args = (exception.message + message,)
            raise exception
        finally:
            self.container._unsaved_statements = []

    def postprocess(self, relaxed_domain_mapping: bool, gams_to_gamspy: bool):
        if gams_to_gamspy:
            super().load_records(relaxed_domain_mapping)

        miro.load_miro_symbol_records(self.container)

        if self.model is not None:
            self.parse_listings()
            return self.prepare_summary(self.trace_file)

        return None



from __future__ import annotations

import itertools
import os
import threading
from typing import TYPE_CHECKING, Any, Literal

import gams.transfer as gt
import pandas as pd
from gams.core.gdx import GMS_DT_SET

import gamspy as gp
import gamspy._algebra.condition as condition
import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._symbols.symbol import Symbol
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from collections.abc import Sequence

    from gamspy import Alias, Container
    from gamspy._algebra.expression import Expression
    from gamspy._symbols.implicits.implicit_set import ImplicitSet
    from gamspy._types import EllipsisType, OperableType
    from gamspy.math.misc import MathOp


class SetMixin:
    @property
    def pos(self: Set | Alias) -> Expression:
        """
        Element position in the current set, starting with 1.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set position values")
        >>> attr[i,"position"]  =  i.pos
        >>> attr.records.values.tolist()
        [['seattle', 'position', 1.0], ['san-diego', 'position', 2.0], ['new-york', 'position', 3.0]]

        """
        return expression.Expression(self, ".", "pos")

    @property
    def ord(self: Set | Alias) -> Expression:
        """
        Same as .pos but for ordered sets only.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set order values")
        >>> attr[i,"order"]  =  i.ord
        >>> attr.records.values.tolist()
        [['seattle', 'order', 1.0], ['san-diego', 'order', 2.0], ['new-york', 'order', 3.0]]

        """
        return expression.Expression(self, ".", "ord")

    @property
    def off(self: Set | Alias) -> Expression:
        """
        Element position in the current set minus 1. So .off = .pos - 1

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set off values")
        >>> attr[i,"off"]  =  i.off
        >>> attr.records.values.tolist()
        [['san-diego', 'off', 1.0], ['new-york', 'off', 2.0]]

        """
        return expression.Expression(self, ".", "off")

    @property
    def rev(self: Set | Alias) -> Expression:
        """
        Reverse element position in the current set, so the value for
        the last element is 0, the value for the penultimate is 1, etc.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set reverse values")
        >>> attr[i,"reverse"]  =  i.rev
        >>> attr.records.values.tolist()
        [['seattle', 'reverse', 2.0], ['san-diego', 'reverse', 1.0]]

        """
        return expression.Expression(self, ".", "rev")

    @property
    def uel(self: Set | Alias) -> Expression:
        """
        Element position in the unique element list.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set uel_pos values")
        >>> attr[i,"uel_position"]  =  i.uel
        >>> attr.records.values.tolist()
        [['seattle', 'uel_position', 1.0], ['san-diego', 'uel_position', 2.0], ['new-york', 'uel_position', 3.0]]

        """
        return expression.Expression(self, ".", "uel")

    @property
    def len(self: Set | Alias) -> Expression:
        """
        Length of the set element name (a count of the number of characters).

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set length values")
        >>> attr[i,"length"]  =  i.len
        >>> attr.records.values.tolist()
        [['seattle', 'length', 7.0], ['san-diego', 'length', 9.0], ['new-york', 'length', 8.0]]

        """
        return expression.Expression(self, ".", "len")

    @property
    def tlen(self: Set | Alias) -> Expression:
        """
        Length of the set element text (a count of the number of characters).

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=[("seattle", "Wisconsin"), ("san-diego", ""), ("new-york", " ")], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set element text length values")
        >>> attr[i,"text_length"]  =  i.tlen
        >>> attr.records.values.tolist()
        [['seattle', 'text_length', 9.0], ['new-york', 'text_length', 1.0]]

        """
        return expression.Expression(self, ".", "tlen")

    @property
    def val(self: Set | Alias) -> Expression:
        """
        If a set element is a number, this attribute gives the value of the number.
        For extended range arithmetic symbols, the symbols are reproduced.
        If a set element is a string that is not a number, then this attribute is
        not defined and trying to use it results in an error.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["12", "20", "-13.4"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set val values")
        >>> attr[i,"value"]  =  i.val
        >>> attr.records.values.tolist()
        [['12', 'value', 12.0], ['20', 'value', 20.0], ['-13.4', 'value', -13.4]]

        """
        return expression.Expression(self, ".", "val")

    @property
    def tval(self: Set | Alias) -> Expression:
        """
        If a set element text is a number, this attribute gives the value of the number.
        For extended range arithmetic symbols, the symbols are reproduced.
        If a set element text is a string that is not a number, then this attribute is
        not defined and trying to use it results in an error.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=[("seattle", "12"), ("san-diego", ""), ("new-york", "-13.4")], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"], description = "Set text values")
        >>> attr[i,"text_value"]  =  i.tval
        >>> attr.records.values.tolist()
        [['seattle', 'text_value', 12.0], ['new-york', 'text_value', -13.4]]

        """
        return expression.Expression(self, ".", "tval")

    @property
    def first(self: Set | Alias) -> Expression:
        """
        Returns 1 for the first set element, otherwise 0.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"])
        >>> attr[i,"is_first"]  =  i.first
        >>> attr.records.values.tolist()
        [['seattle', 'is_first', 1.0]]

        """
        return expression.Expression(self, ".", "first")

    @property
    def last(self: Set | Alias) -> Expression:
        """
        Returns 1 for the last set element, otherwise 0.

        Returns
        -------
        Expression

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego", "new-york"], description="canning plants")
        >>> attr = gp.Parameter(m, "attr", domain = [i, "*"])
        >>> attr[i,"is_last"]  =  i.last
        >>> attr.records.values.tolist()
        [['new-york', 'is_last', 1.0]]

        """
        return expression.Expression(self, ".", "last")

    def lag(
        self,
        n: OperableType,
        type: Literal["linear", "circular"] = "linear",
    ) -> ImplicitSet:
        """
        Lag operation shifts the values of a Set or Alias by one to the left

        Parameters
        ----------
        n : OperableType
        type : 'linear' or 'circular', optional

        Returns
        -------
        ImplicitSet

        Raises
        ------
        ValueError
            When type is not circular or linear

        Examples
        --------
        >>> import gamspy as gp
        >>>
        >>> m = gp.Container()
        >>> t = gp.Set(m, name="t", description="time sequence", records=[f"y-{x}" for x in range(1987, 1992)])
        >>> a = gp.Parameter(m, name="a", domain=[t])
        >>> b = gp.Parameter(m, name="b", domain=[t])
        >>> c = gp.Parameter(m, name="c", domain=[t])
        >>> a[t] = 1986 + gp.Ord(t)
        >>> b[t] = -1
        >>> b[t] = a[t.lag(1, "linear")]
        >>> b.records.values.tolist()
        [['y-1988', 1987.0], ['y-1989', 1988.0], ['y-1990', 1989.0], ['y-1991', 1990.0]]
        >>> c[t] = a[t.lag(1, "circular")]
        >>> c.records.values.tolist()
        [['y-1987', 1991.0], ['y-1988', 1987.0], ['y-1989', 1988.0], ['y-1990', 1989.0], ['y-1991', 1990.0]]

        """
        assert isinstance(self, (gp.Set, gp.Alias))
        jump = n if isinstance(n, int) else n.gamsRepr()  # type: ignore

        if type == "circular":
            return implicits.ImplicitSet(
                self, name=self.name, extension=f" -- {jump}"
            )

        if type == "linear":
            return implicits.ImplicitSet(
                self, name=self.name, extension=f" - {jump}"
            )

        raise ValueError("Lag type must be linear or circular")

    def lead(
        self,
        n: OperableType,
        type: Literal["linear", "circular"] = "linear",
    ) -> ImplicitSet:
        """
        Lead shifts the values of a Set or Alias by one to the right

        Parameters
        ----------
        n : OperableType
        type : 'linear' or 'circular', optional

        Returns
        -------
        ImplicitSet

        Raises
        ------
        ValueError
            When type is not circular or linear

        Examples
        --------
        >>> import gamspy as gp
        >>>
        >>> m = gp.Container()
        >>> t = gp.Set(m, name="t", description="time sequence", records=[f"y-{x}" for x in range(1987, 1992)])
        >>> a = gp.Parameter(m, name="a", domain=[t])
        >>> c = gp.Parameter(m, name="c", domain=[t])
        >>> d = gp.Parameter(m, name="d", domain=[t])
        >>> a[t] = 1986 + gp.Ord(t)
        >>> c[t] = -1
        >>> c[t.lead(2, "linear")] = a[t]
        >>> c.records.values.tolist()
        [['y-1987', -1.0], ['y-1988', -1.0], ['y-1989', 1987.0], ['y-1990', 1988.0], ['y-1991', 1989.0]]
        >>> d[t.lead(2, "circular")] = a[t]
        >>> d.records.values.tolist()
        [['y-1987', 1990.0], ['y-1988', 1991.0], ['y-1989', 1987.0], ['y-1990', 1988.0], ['y-1991', 1989.0]]

        """
        assert isinstance(self, (gp.Set, gp.Alias))
        jump = n if isinstance(n, int) else n.gamsRepr()  # type: ignore

        if type == "circular":
            return implicits.ImplicitSet(
                self, name=self.name, extension=f" ++ {jump}"
            )

        if type == "linear":
            return implicits.ImplicitSet(
                self, name=self.name, extension=f" + {jump}"
            )

        raise ValueError("Lead type must be linear or circular")

    def sameAs(self, other: Set | Alias | str) -> MathOp:
        """
        Evaluates to true if this set is identical to the given set or alias, false otherwise.

        Parameters
        ----------
        other : Set | Alias

        Returns
        -------
        MathOp

        Examples
        --------
        >>> import gamspy as gp

        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego"])
        >>> j = gp.Set(m, name="j", records=["new-york", "seattle"])
        >>> attr = gp.Parameter(m, "attr", domain = [i, j])
        >>> attr[i,j]  =  i.sameAs(j)
        >>> attr.records.values.tolist()
        [['seattle', 'seattle', 1.0]]

        """
        assert isinstance(self, (gt.Set, gt.Alias))
        assert isinstance(other, (gt.Set, gt.Alias, str))
        return gp.math.same_as(self, other)


class Set(gt.Set, operable.Operable, Symbol, SetMixin):
    """
    Represents a Set symbol in GAMS.
    https://gamspy.readthedocs.io/en/latest/user/basics/set.html

    Parameters
    ----------
    container : Container
        Container of the set.
    name : str, optional
        Name of the set. Name is autogenerated by default.
    domain : Sequence[Set | Alias | str] | Set | Alias | str, optional
        Domain of the set.
    is_singleton : bool, optional
        Whether the set is a singleton set. Singleton sets cannot contain more than one element.
    records : pd.DataFrame | np.ndarray | list, optional
        Records of the set.
    domain_forwarding : bool, optional
        Whether the set forwards the domain.
    description : str, optional
        Description of the set.
    uels_on_axes : bool
        Assume that symbol domain information is contained in the axes of the given records.
    is_miro_input : bool
        Whether the symbol is a GAMS MIRO input symbol. See: https://gams.com/miro/tutorial.html
    is_miro_output : bool
        Whether the symbol is a GAMS MIRO output symbol. See: https://gams.com/miro/tutorial.html

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2'])

    """

    @classmethod
    def _constructor_bypass(
        cls,
        container: Container,
        name: str,
        domain: Sequence[Set | Alias | str] | Set | Alias | str = [],
        is_singleton: bool = False,
        records: Any | None = None,
        description: str = "",
    ):
        # create new symbol object
        obj = Set.__new__(
            cls,
            container,
            name,
            domain,
            is_singleton,
            records,
            description=description,
        )

        # set private properties directly
        obj._requires_state_check = False
        obj._container = container
        container._requires_state_check = True
        obj._name = name
        obj._domain = domain
        obj._domain_forwarding = False
        obj._description = description

        obj._records = records
        obj._modified = True
        obj._is_singleton = is_singleton

        # typing
        obj._gams_type = GMS_DT_SET
        obj._gams_subtype = 1 if obj.is_singleton else 0

        # add to container
        container.data.update({name: obj})

        # gamspy attributes
        obj.where = condition.Condition(obj)
        obj.container._add_statement(obj)
        obj._synchronize = True
        obj._metadata = dict()

        # miro support
        obj._is_miro_input = False
        obj._is_miro_output = False

        return obj

    def __new__(
        cls,
        container: Container | None = None,
        name: str | None = None,
        domain: Sequence[Set | Alias | str] | Set | Alias | str | None = None,
        is_singleton: bool = False,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_input: bool = False,
        is_miro_output: bool = False,
    ):
        if container is not None and not isinstance(container, gp.Container):
            raise TypeError(
                "Container must of type `Container` but found"
                f" {type(container)}"
            )

        if name is None:
            return object.__new__(cls)
        else:
            if not isinstance(name, str):
                raise TypeError(
                    f"Name must of type `str` but found {type(name)}"
                )
            try:
                if not container:
                    container = gp._ctx_managers[
                        (os.getpid(), threading.get_native_id())
                    ]

                symbol = container[name]

                if isinstance(symbol, cls):
                    return symbol

                raise TypeError(
                    f"Cannot overwrite symbol `{name}` in container"
                    " because it is not a Set object)"
                )
            except KeyError:
                return object.__new__(cls)

    def __init__(
        self,
        container: Container | None = None,
        name: str | None = None,
        domain: Sequence[Set | Alias | str] | Set | Alias | str | None = None,
        is_singleton: bool = False,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_input: bool = False,
        is_miro_output: bool = False,
    ):
        self._metadata: dict[str, Any] = dict()
        if (is_miro_input or is_miro_output) and name is None:
            raise ValidationError("Please specify a name for miro symbols.")

        self._is_miro_input = is_miro_input
        self._is_miro_output = is_miro_output

        self._synchronize = True

        # domain handling
        if domain is None:
            domain = ["*"]

        if isinstance(domain, (gp.Set, gp.Alias, str)):
            domain = [domain]

        if isinstance(domain, gp.math.Dim):
            domain = gp.math._generate_dims(container, domain.dims)  # type: ignore

        # does symbol exist
        has_symbol = False
        if isinstance(getattr(self, "container", None), gp.Container):
            has_symbol = True

        if has_symbol:
            if any(
                d1 != d2
                for d1, d2 in itertools.zip_longest(self._domain, domain)
            ):
                raise ValueError(
                    "Cannot overwrite symbol in container unless symbol"
                    " domains are equal"
                )

            if self.is_singleton != is_singleton:
                raise ValueError(
                    "Cannot overwrite symbol in container unless"
                    " 'is_singleton' is left unchanged"
                )

            if self.domain_forwarding != domain_forwarding:
                raise ValueError(
                    "Cannot overwrite symbol in container unless"
                    " 'domain_forwarding' is left unchanged"
                )

            # reset some properties
            self._requires_state_check = True
            self.container._requires_state_check = True
            if description != "":
                self.description = description

            previous_state = self.container._options.miro_protect
            self.container._options.miro_protect = False
            self.records = None
            self.modified = True

            # only set records if records are provided
            if records is not None:
                self.setRecords(records, uels_on_axes=uels_on_axes)
            self.container._options.miro_protect = previous_state

        else:
            if container is None:
                try:
                    container = gp._ctx_managers[
                        (os.getpid(), threading.get_native_id())
                    ]
                except KeyError as e:
                    raise ValidationError("Set requires a container.") from e
            assert container is not None

            self.where = condition.Condition(self)
            self._assignment: Expression | None = None

            if name is not None:
                name = validation.validate_name(name)
                if is_miro_input or is_miro_output:
                    name = name.lower()  # type: ignore
            else:
                name = "s" + utils._get_unique_name() + "gpauto"

            self._singleton_check(is_singleton, records, domain)
            previous_state = container._options.miro_protect
            container._options.miro_protect = False

            super().__init__(
                container,
                name,
                domain,
                is_singleton,
                domain_forwarding=domain_forwarding,
                description=description,
                uels_on_axes=uels_on_axes,
            )

            if is_miro_input:
                self._already_loaded = False
                container._miro_input_symbols.append(self.name)

            if is_miro_output:
                container._miro_output_symbols.append(self.name)

            validation.validate_container(self, self._domain)
            self.container._add_statement(self)

            if records is not None:
                self.setRecords(records, uels_on_axes=uels_on_axes)
            else:
                self.modified = False
                self.container._synch_with_gams(
                    gams_to_gamspy=self._is_miro_input
                )

            container._options.miro_protect = previous_state

    def _serialize(self) -> dict:
        info = {
            "_domain_forwarding": self.domain_forwarding,
            "_is_miro_input": self._is_miro_input,
            "_is_miro_output": self._is_miro_output,
            "_metadata": self._metadata,
            "_synchronize": self._synchronize,
        }
        if self._assignment is not None:
            info["_assignment"] = self._assignment.getDeclaration()

        return info

    def _deserialize(self, info: dict) -> None:
        # Set attributes
        for key, value in info.items():
            if key == "_assignment":
                left, right = value.split(" = ")
                value = expression.Expression(left, "=", right[:-1])

            setattr(self, key, value)

        # Relink domain symbols
        new_domain = []
        for elem in self._domain:
            if elem == "*":
                new_domain.append(elem)
                continue
            new_domain.append(self.container[elem])

        self.domain = new_domain

    def __len__(self):
        if self.records is not None:
            return len(self.records.index)

        return 0

    def __getitem__(self, indices: Sequence | str) -> implicits.ImplicitSet:
        domain = validation.validate_domain(self, indices)

        return implicits.ImplicitSet(self, name=self.name, domain=domain)

    def __setitem__(
        self,
        indices: Sequence
        | str
        | int
        | implicits.ImplicitSet
        | EllipsisType
        | slice,
        rhs,
    ):
        # self[domain] = rhs
        domain = validation.validate_domain(self, indices)

        if isinstance(rhs, bool):
            rhs = "yes" if rhs is True else "no"  # type: ignore

        statement = expression.Expression(
            implicits.ImplicitSet(self, name=self.name, domain=domain),
            "=",
            rhs,
        )

        statement._validate_definition(utils._unpack(domain))

        self.container._add_statement(statement)
        self._assignment = statement

        self.container._synch_with_gams(gams_to_gamspy=True)
        self._winner = "gams"

    def __repr__(self) -> str:
        return f"Set(name='{self.name}', domain={self.domain})"

    def _singleton_check(
        self,
        is_singleton: bool,
        records: Any | None,
        domain: Sequence[Set | Alias | str],
    ):
        if is_singleton:
            if records is not None and len(records) != 1:
                raise ValidationError(
                    "Singleton set records size must be one."
                )

            if len(domain) != 1:
                raise ValidationError(
                    f"Length of the domain of the singleton set must be 1 but found {len(domain)}"
                )

    @property
    def records(self):
        """
        Records of the Set

        Returns
        -------
        DataFrame

        Examples
        --------
        >>> import gamspy as gp
        >>> import numpy as np
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i")
        >>> i.setRecords(["seattle", "san-diego"])
        >>> i.records.values.tolist()
        [['seattle', ''], ['san-diego', '']]

        """
        return self._records

    @records.setter
    def records(self, records):
        if (
            hasattr(self, "_is_miro_input")
            and self._is_miro_input
            and self.container._options.miro_protect
        ):
            raise ValidationError(
                "Cannot assign to protected miro input symbols. `miro_protect`"
                " attribute of the container can be set to False to allow"
                " assigning to MIRO input symbols"
            )

        if records is not None and not isinstance(records, pd.DataFrame):
            raise TypeError("Symbol 'records' must be type DataFrame")

        # set records
        self._records = records

        self._requires_state_check = True
        self.modified = True

        self.container._requires_state_check = True
        self.container.modified = True

        if self._records is not None and self.domain_forwarding:
            self._domainForwarding()

            # reset state check flags for all symbols in the container
            for symbol in self.container.data.values():
                symbol._requires_state_check = True

    def setRecords(self, records: Any, uels_on_axes: bool = False) -> None:
        """
        Main convenience method to set standard pandas.DataFrame formatted
        records. If uels_on_axes=True setRecords will assume that all domain
        information is contained in the axes of the pandas object – data will be
        flattened (if necessary).

        Parameters
        ----------
        records : Any
        uels_on_axes : bool, optional

        Examples
        --------
        >>> import gamspy as gp
        >>> import numpy as np
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i")
        >>> i.setRecords(["seattle", "san-diego"])
        >>> i.records.values.tolist()
        [['seattle', ''], ['san-diego', '']]

        """

        super().setRecords(records, uels_on_axes)

        self.container._synch_with_gams(gams_to_gamspy=self._is_miro_input)
        self._winner = "python"

    def gamsRepr(self) -> str:
        """
        Representation of this Set in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", domain=["*"], records=['i1','i2'])
        >>> i.gamsRepr()
        'i'

        """
        return self.name

    def getDeclaration(self) -> str:
        """
        Declaration of the Set in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> i.getDeclaration()
        'Set i(*);'

        """
        output = f"Set {self.name}"

        if self.is_singleton:
            output = f"Singleton {output}"

        output += self._get_domain_str(self.domain_forwarding)

        if self.description:
            output += f' "{self.description}"'

        if self.records is None:
            output += " / /"

        output += ";"

        return output

    def getAssignment(self) -> str:
        """
        Latest assignment to the Set in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> i['i1'] = False
        >>> i.getAssignment()
        'i("i1") = no;'

        """
        if self._assignment is None:
            raise ValidationError("Set was not assigned!")

        return self._assignment.getDeclaration()



from __future__ import annotations

from typing import TYPE_CHECKING

import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._symbols as syms
import gamspy.utils as utils
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    import pandas as pd

    from gamspy import Container


class ExtrinsicFunction(operable.Operable):
    """
    Extrinsic function registered by the user. Extrinsic functions
    are limited to 20 scalar arguments and return a scalar value.
    """

    def __init__(self, container: Container, name: str):
        self.container = container
        self.name = name
        self.args: tuple = ()
        self.domain: list = []
        self.dimension = 0

    @property
    def records(self) -> pd.DataFrame | None:
        """
        Evaluates the extrinsic function and returns the resulting records.

        Returns
        -------
        pd.DataFrame | None
        """
        assert self.container is not None
        temp_name = "a" + utils._get_unique_name()
        temp_param = syms.Parameter._constructor_bypass(
            self.container, temp_name, self.domain
        )
        temp_param[...] = self
        del self.container.data[temp_name]
        return temp_param.records

    def toValue(self) -> float | None:
        """
        Convenience method to return expression records as a Python float. Only possible if there is a single record as a result of the expression evaluation.

        Returns
        -------
        float | None
        """
        records = self.records
        if records is not None:
            return records["value"][0]

        return records

    def __eq__(self, other):
        return expression.Expression(self, "=e=", other)

    def __ne__(self, other):
        return expression.Expression(self, "ne", other)

    def __len__(self):
        return len(self.__str__())

    def __call__(self, *args, **kwds) -> ExtrinsicFunction:
        if kwds:
            raise ValidationError(
                "External functions do not accept keyword arguments"
            )

        self.args = args
        return self

    def __str__(self) -> str:
        representation = self.name

        if self.args:
            arg_strs = []
            for arg in self.args:
                arg_str = (
                    str(arg)
                    if isinstance(arg, (int, float, str))
                    else arg.gamsRepr()
                )
                arg_strs.append(arg_str)

            args = ",".join(arg_strs)

            representation = f"{representation}({args})"

        return representation

    def gamsRepr(self) -> str:
        """
        Representation of this ExtrinsicFunction in GAMS language.

        Returns
        -------
        str
        """
        return self.__str__()

    def getDeclaration(self) -> str:
        """
        Declaration of the ExtrinsicFunction in GAMS

        Returns
        -------
        str
        """
        return self.gamsRepr()

    def latexRepr(self) -> str:
        """
        Representation of this ExtrinsicFunction in Latex.

        Returns
        -------
        str
        """
        return self.gamsRepr()


class ExtrinsicLibrary:
    """
    Extrinsic library registered by the user.

    Parameters
    ----------
    container : str
        Container that the extrinsic library was added to.
    path : str
        Path to the shared object that contains the library.
    functions : dict[str, str]
        Functions to be imported from the extrinsic library.
    """

    def __init__(
        self, container: Container, path: str, functions: dict[str, str]
    ):
        self.container = container
        self.name = "ext" + utils._get_unique_name()
        self.path = path
        self.functions = functions

    def getDeclaration(self) -> str:
        """
        Declaration of the ExtrinsicLibrary in GAMS

        Returns
        -------
        str
        """
        library_str = f"$funclibin {self.name} {self.path}\n"

        for gamspy_name, actual_name in self.functions.items():
            library_str += (
                f"function {gamspy_name} / {self.name}.{actual_name} /;\n"
            )

        return library_str

    def __getattr__(self, name):
        if name in self.functions:
            return ExtrinsicFunction(self.container, name)

        return self.__getattribute__(name)



from __future__ import annotations

import io
from collections.abc import Iterable, Sequence
from typing import TYPE_CHECKING, Any, Literal

from gams.transfer._internals import GAMS_SYMBOL_MAX_LENGTH

import gamspy._symbols as symbols
import gamspy._symbols.implicits as implicits
import gamspy.utils as utils
from gamspy._config import get_option
from gamspy._model import Problem, Sense
from gamspy._options import EXECUTION_OPTIONS, MODEL_ATTR_OPTION_MAP, Options
from gamspy._symbols.symbol import Symbol
from gamspy._types import EllipsisType
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Alias, Equation, Model, Parameter, Set, Variable
    from gamspy._algebra.operation import Operation
    from gamspy._symbols.implicits import (
        ImplicitParameter,
        ImplicitSet,
        ImplicitVariable,
    )
    from gamspy._types import EllipsisType

RESERVED_WORDS = (
    "abort",
    "acronym",
    "acronyms",
    "alias",
    "all",
    "and",
    "binary",
    "break",
    "card",
    "continue",
    "diag",
    "display",
    "do",
    "else",
    "elseif",
    "endfor",
    "endif",
    "endloop",
    "endwhile",
    "eps",
    "equation",
    "equations",
    "execute",
    "execute_load",
    "execute_loaddc",
    "execute_loadhandle",
    "execute_loadpoint",
    "execute_unload",
    "execute_unloaddi",
    "execute_unloadidx",
    "file",
    "files",
    "for",
    "free",
    "function",
    "functions",
    "gdxLoad",
    "if",
    "inf",
    "integer",
    "logic",
    "loop",
    "model",
    "models",
    "na",
    "negative",
    "nonnegative",
    "no",
    "not",
    "option",
    "options",
    "or",
    "ord",
    "parameter",
    "parameters",
    "positive",
    "prod",
    "put",
    "put_utility",
    "putclear",
    "putclose",
    "putfmcl",
    "puthd",
    "putheader",
    "putpage",
    "puttitle",
    "puttl",
    "repeat",
    "sameas",
    "sand",
    "scalar",
    "scalars",
    "semicont",
    "semiint",
    "set",
    "sets",
    "singleton",
    "smax",
    "smin",
    "solve",
    "sor",
    "sos1",
    "sos2",
    "sum",
    "system",
    "table",
    "tables",
    "then",
    "undf",
    "until",
    "variable",
    "variables",
    "while",
    "xor",
    "yes",
)


def get_dimension(
    domain: Sequence[Set | Alias | ImplicitSet | str],
) -> int:
    dimension = 0

    for elem in domain:
        if hasattr(elem, "dimension"):
            dimension += elem.dimension  # type: ignore
        else:
            dimension += 1

    return dimension


def get_domain_path(symbol: Set | Alias | ImplicitSet) -> list[str]:
    path: list[str] = []
    domain = symbol

    while domain != "*":
        if type(domain) is str:
            path.insert(0, domain)
        else:
            path.insert(0, domain.name)

        if type(domain) is symbols.Alias:
            path.insert(0, domain.alias_with.name)  # type: ignore

        domain = "*" if type(domain) is str else domain.domain[0]  # type: ignore

    return path


def validate_dimension(
    symbol: Set
    | Parameter
    | Variable
    | Equation
    | ImplicitSet
    | ImplicitParameter
    | ImplicitVariable
    | Operation,
    domain: list[Set | Alias | ImplicitSet | str],
) -> None:
    dimension = get_dimension(domain)

    if dimension != symbol.dimension:
        raise ValidationError(
            f"The `{symbol}` is referenced with"
            f" {'more' if dimension > symbol.dimension else 'less'} indices"
            f" than declared. Declared dimension is {symbol.dimension} but"
            f" given dimension is {dimension}"
        )


def validate_one_dimensional_sets(
    given: Set | Alias | ImplicitSet,
    actual: str | Set | Alias,
):
    if type(given) is implicits.ImplicitSet:
        return

    given_path = get_domain_path(given)

    if (
        type(actual) is symbols.Set
        and actual.name not in given_path
        or (
            type(actual) is symbols.Alias
            and actual.alias_with.name not in given_path  # type: ignore
        )
    ):
        raise ValidationError(
            f"`Given set `{given.name}` is not a valid domain for declared"
            f" domain `{actual.name}`"
        )


def validate_type(domain):
    for given in domain:
        if type(given) not in (
            symbols.Set,
            symbols.Alias,
            symbols.UniverseAlias,
            implicits.ImplicitSet,
            str,
            int,
            EllipsisType,
            slice,
        ):
            raise TypeError(
                "Domain item must be type Set, Alias, ImplicitSet or str but"
                f" found `{type(given)}`"
            )


def _get_ellipsis_range(domain, given_domain):
    start = 0
    end = len(domain)

    for item in given_domain:
        if type(item) is EllipsisType:
            break

        start += 1

    for item in reversed(given_domain):
        if type(item) is EllipsisType:
            break

        end -= 1

    return start, end


def _expand_ellipsis_slice(
    domain: list[Set | Alias | str],
    indices: Sequence[Set | Alias | str | EllipsisType | slice],
) -> Sequence[Set | Alias | str]:
    if len(domain) == 0:
        # If scalar, only correct indexing is [:] or [...]
        if len(indices) != 1:
            raise ValidationError(
                "Scalar values can only be indexed by '[:]' or '[...]'"
            )

        if type(indices[0]) not in (EllipsisType, slice):
            raise ValidationError(
                "Scalar values can only be indexed by '[:]' or '[...]'"
            )

        return []

    if sum(type(item) is EllipsisType for item in indices) > 1:
        raise ValidationError(
            "There cannot be more than one ellipsis in indexing"
        )

    new_domain: list = []
    index = 0
    for item in indices:
        try:
            dimension = item.dimension  # type: ignore
        except AttributeError:
            dimension = 1

        if type(item) is EllipsisType:
            start, end = _get_ellipsis_range(domain, indices)
            new_domain += domain[start:end]
            index = end
        elif type(item) is slice:
            new_domain.append(domain[index])
            index += dimension
        else:
            new_domain.append(item)
            index += dimension

    return new_domain


def validate_domain(
    symbol: Set
    | Parameter
    | Variable
    | Equation
    | ImplicitSet
    | ImplicitParameter
    | ImplicitVariable
    | Operation,
    indices: EllipsisType | slice | Set | Alias | str | Iterable | ImplicitSet,
):
    domain = utils._to_list(indices)
    domain = [str(elem) if type(elem) is int else elem for elem in domain]
    domain = _expand_ellipsis_slice(symbol.domain, domain)  # type: ignore
    if not get_option("VALIDATION") or not get_option("DOMAIN_VALIDATION"):
        return domain

    validate_type(domain)
    validate_container(symbol, domain)
    validate_dimension(symbol, domain)

    offset = 0
    for given in domain:
        try:
            given_dim = given.dimension
        except AttributeError:
            given_dim = 1
        actual = symbol.domain[offset]

        try:
            actual_dim = actual.dimension
        except AttributeError:
            actual_dim = 1

        if actual_dim == 1 and given_dim == 1:
            if type(given) is str:
                try:
                    if (
                        len(actual.records) < 1000
                        and not actual.records.isin([given]).sum().any()
                    ):
                        raise ValidationError(
                            f"Literal index `{given}` was not found in set"
                            f" `{actual.name}`"
                        )
                except AttributeError:
                    ...
            else:
                validate_one_dimensional_sets(given, actual)

        offset += given_dim

    return domain


def validate_container(
    symbol: Set
    | Parameter
    | Variable
    | Equation
    | ImplicitSet
    | ImplicitParameter
    | ImplicitVariable
    | Operation,
    domain: list[str | Set | Alias],
):
    for set in domain:
        if (
            isinstance(set, (symbols.Set, symbols.Alias))
            and set.container != symbol.container
        ):
            raise ValidationError(
                f"`Domain `{set}` must be in the same container"
                f" with `{symbol}`"
            )


def validate_name(word: str) -> str:
    if not get_option("VALIDATION"):
        return word

    if not isinstance(word, str):
        raise TypeError("Symbol name must be type str")

    if word.lower() in RESERVED_WORDS:
        raise ValidationError(
            "Name cannot be one of the reserved words. List of reserved"
            f" words: {RESERVED_WORDS}"
        )

    if word.endswith("gpauto"):
        raise ValidationError(
            "Name cannot end with one of the reserved words. `gpauto` is a reserverd word."
        )

    return word


def validate_model(
    equations: Iterable[Equation],
    matches: dict[
        Equation | Sequence[Equation],
        Variable | Sequence[Variable],
    ]
    | None,
    problem: Problem | str,
    sense: str | Sense,
) -> tuple[Problem, Sense]:
    if not get_option("VALIDATION"):
        if isinstance(problem, str):
            problem = Problem(problem.upper())

        if isinstance(sense, str):
            sense = Sense(sense.upper())

        return problem, sense

    if isinstance(problem, str):
        if problem.upper() not in Problem.values():
            raise ValueError(
                f"Allowed problem types: {Problem.values()} but found"
                f" {problem}."
            )

        problem = Problem(problem.upper())

    if isinstance(sense, str):
        if sense.upper() not in Sense.values():
            raise ValueError(
                f"Allowed sense values: {Sense.values()} but found {sense}."
            )

        sense = Sense(sense.upper())

    if (
        problem not in (Problem.CNS, Problem.MCP)
        and not isinstance(equations, Iterable)
        or any(
            not isinstance(equation, symbols.Equation)
            for equation in equations
        )
    ):
        raise TypeError("`equations` must be an Iterable of Equation objects")

    if matches is not None:
        if not isinstance(matches, dict):
            raise TypeError(
                f"`matches` must be of type dict but found {type(matches)}"
            )

        if any(
            not isinstance(key, (symbols.Equation, Sequence))
            or not isinstance(value, (symbols.Variable, Sequence))
            or (isinstance(key, Sequence) and isinstance(value, Sequence))
            for key, value in matches.items()
        ):
            raise TypeError(
                "Possible syntaxes for the elements of the `matches` dictionary: Equation:Variable, Equation:Sequence[Variable], or Sequence[Equation]:Variable"
            )

    if problem == Problem.EMP:
        for equation in equations:
            assert equation._definition is not None
            symbol_names = equation._definition._find_all_symbols()
            if any(name.endswith("gpauto") for name in symbol_names):
                raise ValidationError(
                    f"{equation.name} contains symbols with autogenerated names. "
                    "EMP models cannot contain a symbol with an autogenerated name. "
                    "Please specify the `name` argument for all symbols you use in "
                    "EMP models."
                )

    return problem, sense


def validate_model_name(name: str) -> str:
    if not get_option("VALIDATION"):
        return name

    if len(name) == 0:
        raise ValueError("Model name must be at least one character.")

    if len(name) > GAMS_SYMBOL_MAX_LENGTH:
        raise ValueError(
            "Model 'name' is too long, "
            f"max is {GAMS_SYMBOL_MAX_LENGTH} characters"
        )

    if name[0] == "_":
        raise ValidationError(
            "Valid GAMS names cannot begin with a '_' character."
        )

    if not all(True if i == "_" else i.isalnum() for i in name):
        raise ValidationError(
            f"`{name}` is an invalid model name. "
            "Model names can only contain alphanumeric characters "
            "(letters and numbers) and the '_' character."
        )

    return name


def validate_solver_args(
    system_directory: str,
    backend: Literal["local", "engine", "neos"],
    solver: str,
    problem: Problem | str,
    options: Options | None,
    output: io.TextIOWrapper | None,
    load_symbols: list[Symbol] | None,
) -> None:
    if not get_option("VALIDATION"):
        return

    # Check validity of options
    if options is not None and not isinstance(options, Options):
        raise TypeError(
            f"`options` must be of type Option but found {type(options)}"
        )

    # Check validity of output
    if output is not None and (
        not hasattr(output, "write") or not hasattr(output, "flush")
    ):
        raise ValidationError(
            "`output` must write and flush operations but found"
            f" {type(output)} which does not support them."
        )

    # Check validity of load_symbols
    if load_symbols is not None:
        if not isinstance(load_symbols, list):
            raise ValidationError(
                f"`load_symbols` must be list of Symbol objects. Given type: {type(load_symbols)}"
            )

        for elem in load_symbols:
            if not isinstance(elem, Symbol):
                raise ValidationError(
                    f"Elements of `load_symbols` must be of type Symbol but found {elem}"
                )

    # Check validity of solver
    if not isinstance(solver, str):
        raise TypeError("`solver` argument must be a string.")

    if backend == "neos" and solver.lower() in ("mpsge", "kestrel"):
        raise ValidationError(
            f"`{solver}` is not a valid solver for NEOS Server."
        )

    if backend == "engine" and solver.lower() == "mpsge":
        raise ValidationError(
            f"`{solver}` is not a valid solver for GAMS Engine."
        )

    if backend == "local":
        # No need to check whether the solver is installed on client's machine for NEOS or ENGINE.
        installed_solvers = utils.getInstalledSolvers(system_directory)
        if solver.upper() not in installed_solvers:
            raise ValidationError(
                f"Provided solver name `{solver}` is not installed on your"
                f" machine. Install `{solver}` with `gamspy install solver"
                f" {solver.lower()}`"
            )

        capabilities = utils.getSolverCapabilities(system_directory)
        if str(problem).upper() not in capabilities[solver.upper()]:
            raise ValidationError(
                f"Given solver `{solver}` is not capable of solving given"
                f" problem type `{problem}`. See capability matrix "
                "(https://www.gams.com/latest/docs/S_MAIN.html#SOLVERS_MODEL_TYPES)"
                " to choose a suitable solver"
            )


def validate_equations(model: Model) -> None:
    if not get_option("VALIDATION"):
        return

    if model.container._is_restarted:
        return

    for equation in model.equations:
        if equation._definition is None:
            if equation.dimension == 0:
                name = equation.name
                raise ValidationError(
                    f"`{name}` has been declared as a scalar equation but no equation definition was found.\n"
                    f"The definition of `{name}` must use either `[:]` or `[...]` after the python variable name for the definition to register. "
                    f"For example: \n\n\t{name} = gp.Equation(..., name='{name}', ...)\n\t{name}[...] = LHS == RHS\n\nFailure to add `[:]` or `[...]`, like in the following:\n\n\t"
                    f"{name} = LHS == RHS\n\njust redefines the Python variable `{name}` as an expression, but does not define the equation `{name}`. "
                    "This issue stems from the lack of assignment operator overloading in Python.\n"
                    f"You can verify that the equation `{name}` has been defined by printing its definition:\n\n\t"
                    f"print({name}.getDefinition())"
                )

            raise ValidationError(
                f"`{equation.name}` has been declared as an equation but no equation definition was found."
            )


def validate_global_options(options: Any) -> Options:
    if not get_option("VALIDATION"):
        if options is None:
            return Options()

        return options

    if options is not None and not isinstance(options, Options):
        raise TypeError(
            f"`options` must be of type Option but found {type(options)}"
        )

    if isinstance(options, Options):
        options_dict = options.model_dump(exclude_none=True)
        if any(option in options_dict for option in MODEL_ATTR_OPTION_MAP):
            raise ValidationError(
                f"The following model options cannot be provided at Container creation time: {', '.join(MODEL_ATTR_OPTION_MAP.keys())}."
            )

        if any(option in options_dict for option in EXECUTION_OPTIONS):
            raise ValidationError(
                f"The following runtime options cannot be provided at Container creation time: {', '.join(EXECUTION_OPTIONS.keys())}."
            )

    if options is None:
        return Options()

    return options



from __future__ import annotations

from typing import Union

from typing_extensions import TypeAlias

from gamspy._algebra.operable import Operable

OperableType: TypeAlias = Union[Operable, int, float]
EllipsisType: TypeAlias = type(...)  # type: ignore



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

import gamspy as gp
import gamspy._algebra.condition as condition
import gamspy.utils as utils
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Alias, Product, Sand, Set, Smax, Smin, Sor, Sum
    from gamspy._symbols.implicits import (
        ImplicitParameter,
        ImplicitSet,
        ImplicitVariable,
    )


class ImplicitSymbol(ABC):
    def __init__(
        self,
        parent,
        name,
        domain,
        permutation=None,
        parent_scalar_domains=None,
    ) -> None:
        self.parent = parent
        self.container = parent.container
        self.name = name
        self.domain = domain
        self.where = condition.Condition(self)

        if permutation is not None:
            permutation = list(permutation)

        self.permutation = permutation

        if parent_scalar_domains is None:
            parent_scalar_domains = []
        else:
            parent_scalar_domains = list(parent_scalar_domains)

        self.parent_scalar_domains = parent_scalar_domains

        self.fix_domain_scalars(parent_scalar_domains)

    def __bool__(self):
        raise ValidationError("A symbol cannot be used as a truth value.")

    def fix_domain_scalars(self, parent_scalar_domains):
        if len(self.domain) == 1 and self.domain[0] == "*":
            self._scalar_domains = []
            return

        bare_domain = utils._get_set(self.domain)
        domain = []
        scalars = []
        permutation_indices_to_del = []
        for i, d in enumerate(bare_domain):
            if isinstance(d, str):
                loc_scalar = (
                    i if self.permutation is None else self.permutation[i]
                )
                scalars.append((loc_scalar, d))
                permutation_indices_to_del.append(i)
            else:
                domain.append(d)

        permutation_indices_to_del = reversed(permutation_indices_to_del)
        if self.permutation is not None:
            for index in permutation_indices_to_del:
                del self.permutation[index]

            self.fix_permutation()

        self.domain = domain

        scalars.extend(parent_scalar_domains)
        scalars = list(sorted(scalars, key=lambda k: k[0]))
        self._scalar_domains = scalars

    def fix_permutation(self):
        numbers_pos = list(zip(self.permutation, range(len(self.permutation))))
        numbers_pos = sorted(numbers_pos, key=lambda k: k[0])
        new_perm = [-1] * len(self.permutation)
        for i, (_, pos) in enumerate(numbers_pos):
            new_perm[pos] = i

        self.permutation = new_perm

    @property
    def dimension(self) -> int:
        return self.parent.dimension - len(self._scalar_domains)

    @abstractmethod
    def gamsRepr(self):
        """Representation of the implicit symbol in GAMS"""

    def latexRepr(self):
        from .implicit_set import ImplicitSet

        name = self.name.replace("_", "\\_")
        representation = f"{name}"
        domain = list(self.domain)

        for i, d in self._scalar_domains:
            domain.insert(i, d)

        if domain:
            set_strs = []
            for elem in domain:
                if isinstance(elem, (gp.Set, gp.Alias, ImplicitSet)):
                    set_strs.append(f"\\text{{{elem.latexRepr()}}}")
                elif isinstance(elem, str):
                    set_strs.append(
                        f"\\text{{\\textquotesingle {elem}\\textquotesingle}}"
                    )

            domain_str = ",".join(set_strs)
            representation = f"{representation}_\\text{{{domain_str}}}"

        return representation

    def sum(
        self: ImplicitSet | ImplicitParameter | ImplicitVariable,
        *indices: Set | Alias,
    ) -> Sum:
        """
        Equivalent to Sum(indices, obj[obj.domain]). For example:

        v = Variable(m, domain=[i, j, k])
        v.sum() is equivalent to Sum((i,j), v[i, j, k])
        v.sum(i) is equivalent to Sum(i, v[i, j, k])
        v.sum(i, j) is equivalent to Sum((i, j), v[i, j, k])

        Returns
        -------
        Sum
            Generated Sum gp.

        Raises
        ------
        ValidationError
            In case the symbol is scalar.
        """
        if not self.domain:
            raise ValidationError(
                "Sum operation is not possible on scalar parameters."
            )

        if not indices:
            indices = self.domain

        return gp.Sum(indices, self[self.domain])

    def product(
        self: ImplicitSet | ImplicitParameter | ImplicitVariable,
        *indices: Set | Alias,
    ) -> Product:
        """
        Equivalent to Product(indices, obj[obj.domain]). For example:

        v = Variable(m, domain=[i, j, k])
        v.product() is equivalent to Product((i,j), v[i, j, k])
        v.product(i) is equivalent to Product(i, v[i, j, k])
        v.product(i, j) is equivalent to Product((i, j), v[i, j, k])

        Returns
        -------
        Product
            Generated Product gp.

        Raises
        ------
        ValidationError
            In case the symbol is scalar.
        """
        if not self.domain:
            raise ValidationError(
                "Product operation is not possible on scalar parameters."
            )

        if not indices:
            indices = self.domain

        return gp.Product(indices, self[self.domain])

    def smin(
        self: ImplicitSet | ImplicitParameter | ImplicitVariable,
        *indices: Set | Alias,
    ) -> Smin:
        """
        Equivalent to Smin(indices, obj[obj.domain]). For example:

        v = Variable(m, domain=[i, j, k])
        v.smin() is equivalent to Smin((i,j), v[i, j, k])
        v.smin(i) is equivalent to Smin(i, v[i, j, k])
        v.smin(i, j) is equivalent to Smin((i, j), v[i, j, k])

        Returns
        -------
        Smin
            Generated Smin gp.

        Raises
        ------
        ValidationError
            In case the symbol is scalar.
        """
        if not self.domain:
            raise ValidationError(
                "Smin operation is not possible on scalar parameters."
            )

        if not indices:
            indices = self.domain

        return gp.Smin(indices, self[self.domain])

    def smax(
        self: ImplicitSet | ImplicitParameter | ImplicitVariable,
        *indices: Set | Alias,
    ) -> Smax:
        """
        Equivalent to Smax(indices, obj[obj.domain]). For example:

        v = Variable(m, domain=[i, j, k])
        v.smax() is equivalent to Smax((i,j), v[i, j, k])
        v.smax(i) is equivalent to Smax(i, v[i, j, k])
        v.smax(i, j) is equivalent to Smax((i, j), v[i, j, k])

        Returns
        -------
        Smax
            Generated Smax gp.

        Raises
        ------
        ValidationError
            In case the symbol is scalar.
        """
        if not self.domain:
            raise ValidationError(
                "Smax operation is not possible on scalar parameters."
            )

        if not indices:
            indices = self.domain

        return gp.Smax(indices, self[self.domain])

    def sand(
        self: ImplicitSet | ImplicitParameter | ImplicitVariable,
        *indices: Set | Alias,
    ) -> Sand:
        """
        Equivalent to Sand(indices, obj[obj.domain]). For example:

        v = Variable(m, domain=[i, j, k])
        v.sand() is equivalent to Sand((i,j), v[i, j, k])
        v.sand(i) is equivalent to Sand(i, v[i, j, k])
        v.sand(i, j) is equivalent to Sand((i, j), v[i, j, k])

        Returns
        -------
        Sand
            Generated Sand gp.

        Raises
        ------
        ValidationError
            In case the symbol is scalar.
        """
        if not self.domain:
            raise ValidationError(
                "Sand operation is not possible on scalar parameters."
            )

        if not indices:
            indices = self.domain

        return gp.Sand(indices, self[self.domain])

    def sor(
        self: ImplicitSet | ImplicitParameter | ImplicitVariable,
        *indices: Set | Alias,
    ) -> Sor:
        """
        Equivalent to Sor(indices, obj[obj.domain]). For example:

        v = Variable(m, domain=[i, j, k])
        v.sor() is equivalent to Sor((i,j), v[i, j, k])
        v.sor(i) is equivalent to Sor(i, v[i, j, k])
        v.sor(i, j) is equivalent to Sor((i, j), v[i, j, k])

        Returns
        -------
        Sor
            Generated Sor gp.

        Raises
        ------
        ValidationError
            In case the symbol is scalar.
        """
        if not self.domain:
            raise ValidationError(
                "Sor operation is not possible on scalar parameters."
            )

        if not indices:
            indices = self.domain

        return gp.Sor(indices, self[self.domain])



from __future__ import annotations

import gamspy._algebra.condition as condition
import gamspy._algebra.operable as operable
from gamspy._algebra import expression


class Number(operable.Operable):
    """
    Needed for conditions on numbers.

    Parameters
    ----------
    value : int | float

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> k = gp.Set(m, "k", records=["1964-i","1964-ii","1964-iii","1964-iv"])
    >>> ki = gp.Set(m, name="ki", domain=[k], description="initial period")
    >>> ki[k] = gp.Number(1).where[gp.Ord(k) == 1]

    """

    def __init__(self, value: int | float):
        if not isinstance(value, (int, float)):
            raise TypeError(
                f"Number should be of type int or float but given value has type `{type(value)}`"
            )

        self._value = value
        self.where = condition.Condition(self)
        self.domain: list = []

    def __eq__(self, other):
        return expression.Expression(self, "=e=", other)

    def __ne__(self, other):
        return expression.Expression(self, "ne", other)

    def __repr__(self) -> str:
        return f"Number(value={self._value})"

    def gamsRepr(self) -> str:
        """
        Representation of this Number in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> gp.Number(2.0).gamsRepr()
        '2.0'

        """
        return f"{self._value}"

    def latexRepr(self) -> str:
        """
        Representation of this Number in Latex.

        Returns
        -------
        str
        """
        return f"{self._value}"



from __future__ import annotations

import itertools
import os
import threading
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

import gams.transfer as gt
import numpy as np
import pandas as pd
from gams.core.gdx import GMS_DT_PAR

import gamspy as gp
import gamspy._algebra.condition as condition
import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._algebra.operation as operation
import gamspy._symbols.implicits as implicits
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._symbols.symbol import Symbol
from gamspy.exceptions import ValidationError

if TYPE_CHECKING:
    from gamspy import Alias, Container, Set
    from gamspy._algebra.expression import Expression
    from gamspy._algebra.operation import Operation
    from gamspy._symbols.implicits import ImplicitParameter
    from gamspy._types import EllipsisType
    from gamspy.math.matrix import Dim


class Parameter(gt.Parameter, operable.Operable, Symbol):
    """
    Represents a parameter symbol in GAMS.
    https://gamspy.readthedocs.io/en/latest/user/basics/parameter.html

    Parameters
    ----------
    container : Container
        Container of the parameter.
    name : str, optional
        Name of the parameter. Name is autogenerated by default.
    domain : Sequence[Set | Alias | str] | Set | Alias | str, optional
        Domain of the parameter.
    records : int | float | pd.DataFrame | np.ndarray | list, optional
        Records of the parameter.
    domain_forwarding : bool, optional
        Whether the parameter forwards the domain.
    description : str, optional
        Description of the parameter.
    uels_on_axes : bool
        Assume that symbol domain information is contained in the axes of the given records.
    is_miro_input : bool
        Whether the symbol is a GAMS MIRO input symbol. See: https://gams.com/miro/tutorial.html
    is_miro_output : bool
        Whether the symbol is a GAMS MIRO output symbol. See: https://gams.com/miro/tutorial.html
    is_miro_table : bool
        Whether the symbol is a GAMS MIRO table symbol. See: https://gams.com/miro/tutorial.html

    Examples
    --------
    >>> import gamspy as gp
    >>> m = gp.Container()
    >>> i = gp.Set(m, "i", records=['i1','i2'])
    >>> a = gp.Parameter(m, "a", [i], records=[['i1',1],['i2',2]])

    """

    @classmethod
    def _constructor_bypass(
        cls,
        container: Container,
        name: str,
        domain: Sequence[Set | Alias | str] | Set | Alias | Dim | str = [],
        records: Any | None = None,
        description: str = "",
    ):
        obj = Parameter.__new__(
            cls, container, name, domain, records, description=description
        )

        # set private properties directly
        obj._requires_state_check = False
        obj._container = container
        container._requires_state_check = True
        obj._name = name
        obj._domain = domain
        obj._domain_forwarding = False
        obj._description = description
        obj._records = records
        obj._modified = True

        # typing
        obj._gams_type = GMS_DT_PAR
        obj._gams_subtype = 0

        # add to container
        container.data.update({name: obj})

        # gamspy attributes
        obj._synchronize = True
        obj.where = condition.Condition(obj)
        obj.container._add_statement(obj)
        obj._metadata = dict()

        # miro support
        obj._is_miro_input = False
        obj._is_miro_output = False
        obj._is_miro_table = False

        return obj

    def __new__(
        cls,
        container: Container | None = None,
        name: str | None = None,
        domain: Sequence[Set | Alias | str]
        | Set
        | Alias
        | Dim
        | str
        | None = None,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_input: bool = False,
        is_miro_output: bool = False,
        is_miro_table: bool = False,
    ):
        if container is not None and not isinstance(container, gp.Container):
            raise TypeError(
                "Container must of type `Container` but found"
                f" {type(container)}"
            )

        if name is None:
            return object.__new__(cls)
        else:
            if not isinstance(name, str):
                raise TypeError(
                    f"Name must of type `str` but found {type(name)}"
                )

            try:
                if not container:
                    container = gp._ctx_managers[
                        (os.getpid(), threading.get_native_id())
                    ]

                symbol = container[name]
                if isinstance(symbol, cls):
                    return symbol

                raise TypeError(
                    f"Cannot overwrite symbol `{name}` in container"
                    " because it is not a Parameter object"
                )
            except KeyError:
                return object.__new__(cls)

    def __init__(
        self,
        container: Container | None = None,
        name: str | None = None,
        domain: Sequence[Set | Alias | str]
        | Set
        | Alias
        | Dim
        | str
        | None = None,
        records: Any | None = None,
        domain_forwarding: bool = False,
        description: str = "",
        uels_on_axes: bool = False,
        is_miro_input: bool = False,
        is_miro_output: bool = False,
        is_miro_table: bool = False,
    ):
        self._metadata: dict[str, Any] = dict()
        if (is_miro_input or is_miro_output) and name is None:
            raise ValidationError("Please specify a name for miro symbols.")

        # miro support
        self._is_miro_input = is_miro_input
        self._is_miro_output = is_miro_output
        self._is_miro_table = is_miro_table

        self._synchronize = True

        # domain handling
        if domain is None:
            domain = []

        if isinstance(domain, (gp.Set, gp.Alias, str)):
            domain = [domain]

        if isinstance(domain, gp.math.Dim):
            domain = gp.math._generate_dims(container, domain.dims)  # type: ignore

        # does symbol exist
        has_symbol = False
        if isinstance(getattr(self, "container", None), gp.Container):
            has_symbol = True

        if has_symbol:
            if any(
                d1 != d2
                for d1, d2 in itertools.zip_longest(self._domain, domain)
            ):
                raise ValueError(
                    "Cannot overwrite symbol in container unless symbol"
                    " domains are equal"
                )

            if self.domain_forwarding != domain_forwarding:
                raise ValueError(
                    "Cannot overwrite symbol in container unless"
                    " 'domain_forwarding' is left unchanged"
                )

            # reset some properties
            self._requires_state_check = True
            self.container._requires_state_check = True
            if description != "":
                self.description = description

            previous_state = self.container._options.miro_protect
            self.container._options.miro_protect = False
            self.records = None
            self.modified = True

            # only set records if records are provided
            if records is not None:
                self.setRecords(records, uels_on_axes=uels_on_axes)
            self.container._options.miro_protect = previous_state
        else:
            if container is None:
                try:
                    container = gp._ctx_managers[
                        (os.getpid(), threading.get_native_id())
                    ]
                except KeyError as e:
                    raise ValidationError(
                        "Parameter requires a container."
                    ) from e
            assert container is not None

            if name is not None:
                name = validation.validate_name(name)

                if is_miro_input or is_miro_output:
                    name = name.lower()  # type: ignore
            else:
                name = "p" + utils._get_unique_name() + "gpauto"

            previous_state = container._options.miro_protect
            container._options.miro_protect = False
            super().__init__(
                container,
                name,
                domain,
                domain_forwarding=domain_forwarding,
                description=description,
                uels_on_axes=uels_on_axes,
            )

            if is_miro_input:
                self._already_loaded = False
                container._miro_input_symbols.append(self.name)

            if is_miro_output:
                container._miro_output_symbols.append(self.name)

            validation.validate_container(self, self._domain)
            self.where = condition.Condition(self)
            self._assignment: Expression | None = None
            self.container._add_statement(self)

            if records is not None:
                super().setRecords(records, uels_on_axes=uels_on_axes)
                if self.dimension == 0:
                    self.modified = False
            else:
                self.modified = False

            self.container._synch_with_gams(gams_to_gamspy=self._is_miro_input)

            container._options.miro_protect = previous_state

    def _serialize(self) -> dict:
        info = {
            "_domain_forwarding": self.domain_forwarding,
            "_is_miro_input": self._is_miro_input,
            "_is_miro_output": self._is_miro_output,
            "_is_miro_table": self._is_miro_table,
            "_metadata": self._metadata,
            "_synchronize": self._synchronize,
        }
        if self._assignment is not None:
            info["_assignment"] = self._assignment.getDeclaration()

        return info

    def _deserialize(self, info: dict) -> None:
        for key, value in info.items():
            if key == "_assignment":
                left, right = value.split(" = ")
                value = expression.Expression(left, "=", right[:-1])

            setattr(self, key, value)

        # Relink domain symbols
        new_domain = []
        for elem in self._domain:
            if elem == "*":
                new_domain.append(elem)
                continue
            new_domain.append(self.container[elem])

        self.domain = new_domain

    def __getitem__(
        self, indices: EllipsisType | slice | Sequence | str
    ) -> implicits.ImplicitParameter:
        domain = validation.validate_domain(self, indices)

        return implicits.ImplicitParameter(self, name=self.name, domain=domain)

    def __setitem__(
        self,
        indices: EllipsisType | slice | Sequence | str | implicits.ImplicitSet,
        rhs: Operation | Expression | ImplicitParameter | float | int,
    ):
        # self[domain] = rhs
        domain = validation.validate_domain(self, indices)

        if self._is_miro_input and self.container._options.miro_protect:
            raise ValidationError(
                f"Cannot assign to protected miro input symbol {self.name}. `miro_protect`"
                " attribute of the container can be set to False to allow"
                " assigning to MIRO input symbols"
            )

        if isinstance(rhs, float):
            rhs = utils._map_special_values(rhs)  # type: ignore

        statement = expression.Expression(
            implicits.ImplicitParameter(self, name=self.name, domain=domain),
            "=",
            rhs,
        )

        statement._validate_definition(utils._unpack(domain))

        self.container._add_statement(statement)
        self._assignment = statement

        self.container._synch_with_gams(gams_to_gamspy=True)
        self._winner = "gams"

    def __eq__(self, other):
        op = "eq"
        if isinstance(
            other,
            (
                implicits.ImplicitVariable,
                expression.Expression,
                operation.Operation,
            ),
        ):
            op = "=e="
        return expression.Expression(self, op, other)

    def __ne__(self, other):
        return expression.Expression(self, "ne", other)

    def __neg__(self):
        return expression.Expression(None, "-", self)

    def __repr__(self) -> str:
        return f"Parameter(name='{self.name}', domain={self.domain})"

    @property
    def T(self) -> implicits.ImplicitParameter:
        """See gamspy.Parameter.t"""
        return self.t()

    def t(self) -> implicits.ImplicitParameter:
        """Returns an ImplicitParameter derived from this
        parameter by swapping its last two indices. This operation
        does not generate a new parameter in GAMS.

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> j = gp.Set(m, "j", records=['j1','j2'])
        >>> v = gp.Parameter(m, "v", domain=[i, j])
        >>> v_t = v.t()
        >>> v_t.domain
        [Set(name='j', domain=['*']), Set(name='i', domain=['*'])]
        >>> v_t[i, j] # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        gamspy.exceptions.ValidationError:
        >>> v_t["j1", "i1"].gamsRepr()
        'v("i1","j1")'

        """
        from gamspy.math.matrix import permute

        dims = [x for x in range(len(self.domain))]
        if len(dims) < 2:
            raise ValidationError(
                "Parameter must contain at least 2 dimensions to transpose"
            )

        x = dims[-1]
        dims[-1] = dims[-2]
        dims[-2] = x
        return permute(self, dims)  # type: ignore

    @property
    def records(self):
        """
        Records of the Parameter

        Returns
        -------
        DataFrame

        Examples
        --------
        >>> import gamspy as gp
        >>> import numpy as np
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i", records=["seattle", "san-diego"])
        >>> d = gp.Parameter(m, name="d", domain=[i])
        >>> d.setRecords(np.array([10, 25]))
        >>> d.records.values.tolist()
        [['seattle', 10.0], ['san-diego', 25.0]]

        """
        return self._records

    @records.setter
    def records(self, records):
        if (
            hasattr(self, "_is_miro_input")
            and self._is_miro_input
            and self.container._options.miro_protect
        ):
            raise ValidationError(
                "Cannot assign to protected miro input symbols. `miro_protect`"
                " attribute of the container can be set to False to allow"
                " assigning to MIRO input symbols"
            )

        if records is not None and not isinstance(records, pd.DataFrame):
            raise TypeError("Symbol 'records' must be type DataFrame")

        # set records
        self._records = records

        self._requires_state_check = True
        self.modified = True

        self.container._requires_state_check = True
        self.container.modified = True

        if self._records is not None and self.domain_forwarding:
            self._domainForwarding()

            # reset state check flags for all symbols in the container
            for symbol in self.container.data.values():
                symbol._requires_state_check = True

    def setRecords(self, records: Any, uels_on_axes: bool = False) -> None:
        """
        Main convenience method to set standard pandas.DataFrame formatted
        records. If uels_on_axes=True setRecords will assume that all domain
        information is contained in the axes of the pandas object – data will be
        flattened (if necessary).

        Parameters
        ----------
        records : Any
        uels_on_axes : bool, optional

        Examples
        --------
        >>> import gamspy as gp
        >>> import numpy as np
        >>> m = gp.Container()
        >>> i = gp.Set(m, name="i")
        >>> i.setRecords(["seattle", "san-diego"])
        >>> i.records.values.tolist()
        [['seattle', ''], ['san-diego', '']]

        """
        super().setRecords(records, uels_on_axes)

        self.container._synch_with_gams(gams_to_gamspy=self._is_miro_input)
        self._winner = "python"

    def gamsRepr(self) -> str:
        """
        Representation of this Parameter in GAMS language.

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> d = gp.Parameter(m, "d", domain=i)
        >>> d.gamsRepr()
        'd(i)'

        """
        representation = self.name
        if self.domain:
            representation += self._get_domain_str(self.domain_forwarding)

        return representation

    def getDeclaration(self) -> str:
        """
        Declaration of the Parameter in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> a = gp.Parameter(m, "a", [i], records=[['i1',1],['i2',2]])
        >>> a.getDeclaration()
        'Parameter a(i);'

        """
        output = f"Parameter {self.gamsRepr()}"

        if self.description:
            output += ' "' + self.description + '"'

        if self.records is None:
            output += " / /"

        if self.dimension == 0 and self.records is not None:
            value = self.toValue()
            value = utils._map_special_values(value)
            if isinstance(value, float) and np.isnan(value):
                value = "Undf"

            output += f" / {value} /"

        output += ";"

        return output

    def getAssignment(self) -> str:
        """
        Latest assignment to the Parameter in GAMS

        Returns
        -------
        str

        Examples
        --------
        >>> import gamspy as gp
        >>> m = gp.Container()
        >>> i = gp.Set(m, "i", records=['i1','i2'])
        >>> a = gp.Parameter(m, "a", [i], records=[['i1',1],['i2',2]])
        >>> a[i] = a[i] * 5
        >>> a.getAssignment()
        'a(i) = (a(i) * 5);'

        """
        if self._assignment is None:
            raise ValidationError("Parameter was not assigned!")

        return self._assignment.getDeclaration()



from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

import gamspy._algebra.expression as expression
import gamspy._algebra.operable as operable
import gamspy._validation as validation
import gamspy.utils as utils
from gamspy._symbols.implicits.implicit_symbol import ImplicitSymbol

if TYPE_CHECKING:
    import pandas as pd

    from gamspy import Alias, Set
    from gamspy._algebra.expression import Expression
    from gamspy._types import EllipsisType


class ImplicitSet(ImplicitSymbol, operable.Operable):
    """
    Implicit Set

    Parameters
    ----------
    container : Container
    name : str
    domain : list[Set | str], optional
    """

    def __init__(
        self,
        parent: Set | Alias,
        name: str,
        domain: list[Set | str] = ["*"],
        scalar_domains: list[tuple[int, Set]] | None = None,
        extension: str | None = None,
    ) -> None:
        super().__init__(
            parent, name, domain, parent_scalar_domains=scalar_domains
        )
        self.extension = extension

    def __ge__(self, other) -> Expression:
        return expression.Expression(self, ">=", other)

    def __le__(self, other) -> Expression:
        return expression.Expression(self, "<=", other)

    def __repr__(self) -> str:
        return f"ImplicitSet(parent={self.parent}, name='{self.name}', domain={self.domain}, extension={self.extension}, parent_scalar_domains={self.parent_scalar_domains})"

    def __getitem__(
        self, indices: Sequence | str | EllipsisType | slice
    ) -> ImplicitSet:
        domain = validation.validate_domain(self, indices)
        return ImplicitSet(
            parent=self.parent,
            name=self.name,
            domain=domain,
            scalar_domains=self._scalar_domains,
        )

    @property
    def records(self) -> pd.DataFrame | None:
        if self.parent.records is None:
            return None

        recs = self.parent.records
        for idx, literal in self._scalar_domains:
            column_name = recs.columns[idx]
            recs = recs[recs[column_name] == literal]

        return recs

    def gamsRepr(self) -> str:
        representation = self.name

        if self.extension is not None:
            representation += f"{self.extension}"

        if self.domain != ["*"]:
            domain = list(self.domain)
            for i, d in self._scalar_domains:
                domain.insert(i, d)

            representation += utils._get_domain_str(domain)

        return representation



from __future__ import annotations

from typing import TYPE_CHECKING

import gamspy._symbols.implicits as implicits
import gamspy.math
import gamspy.utils as utils
from gamspy._symbols.parameter import Parameter
from gamspy._symbols.variable import Variable
from gamspy.exceptions import ValidationError
from gamspy.math.matrix import next_alias

if TYPE_CHECKING:
    from gamspy._algebra.expression import Expression
    from gamspy._algebra.operation import Operation


def _get_random_name(prefix: str) -> str:
    return f"{prefix}_{utils._get_unique_name()}"


def _get_lb(x: Variable, default_lb: float):
    lb = x.container.addParameter(
        _get_random_name("lb"),
        domain=x.domain,
    )
    lb[...] = x.lo[...]
    lb[...].where[lb[...] == "-inf"] = default_lb
    return lb


def _get_ub(x: Variable, default_ub: float):
    ub = x.container.addParameter(
        _get_random_name("ub"),
        domain=x.domain,
    )
    ub[...] = x.up[...]
    ub[...].where[ub[...] == "inf"] = default_ub
    return ub


def relu_with_sos1_var(
    x: (
        Parameter
        | Variable
        | implicits.ImplicitParameter
        | implicits.ImplicitVariable
        | Expression
        | Operation
    ),
    return_slack_var: bool = False,
):
    """
    Implements the ReLU activation function using
    `SOS1 <https://www.gams.com/47/docs/UG_LanguageFeatures.html?search=sos#UG_LanguageFeatures_SpecialOrderSetsOfType1-SOS1>`_ variables.
    The ReLU function is defined as ReLU(x) = max(x, 0). This implementation
    **generates** one SOS1 type variable which is necessary to represent the
    mathematical relationship and one equation. The SOS1 variable contains the
    activation variable and the slack variable.

    Unlike ``relu_with_binary_var``, this function does not require lower and
    upper bounds for the formulation. It is claimed that when providing tight
    bounds is not straightforward, using ``relu_with_sos1_var`` might perform
    better than ``relu_with_binary_var``, as the relaxation of
    ``relu_with_binary_var`` can be weak.

    Usage of SOS1 variables require MIP or MINLP and a solver that supports SOS1
    variables. Main intended use case of this function is embedding the trained
    neural network into MIP models, we do not suggest using it in training since
    you would need a MINLP solver that support SOS1 variables.

    Returns the activation variable and the equation list if ``return_slack_var`` is
    False, otherwise returns activation, slack variable and the equation list in
    order. Since activation variable and slack variable are the same variable
    only separated by the last domain this function returns ImplicitVariable
    instead of Variable.

    Based on paper:
    `PySCIPOpt-ML: Embedding trained machine learning models into mixed-integer programs. <https://arxiv.org/pdf/2312.08074>`_

    Parameters
    ----------
    x : Parameter | Variable | implicits.ImplicitParameter | implicits.ImplicitVariable | Expression | Operation
    return_slack_var: bool

    Returns
    -------
    tuple[implicits.ImplicitVariable, list[Equation]] | tuple[implicits.ImplicitVariable, implicits.ImplicitVariable, list[Equation]]

    Examples
    --------
    >>> from gamspy import Container, Variable, Set
    >>> from gamspy.math.activation import relu_with_sos1_var
    >>> m = Container()
    >>> i = Set(m, "i", records=range(3))
    >>> x = Variable(m, "x", domain=[i])
    >>> y, eqs = relu_with_sos1_var(x)
    >>> y, s, eqs = relu_with_sos1_var(x, return_slack_var=True)
    >>> y.domain # implicit activation variable has the same domain
    [Set(name='i', domain=['*'])]
    >>> s.domain # implicit slack variable has the same domain as well
    [Set(name='i', domain=['*'])]
    >>> y.name == s.name # In the background that y and s are parts of the same variable
    True
    >>> id(y.parent) == id(s.parent)
    True

    """
    domain = x.domain

    assert x.container
    last_dim = gamspy.math._generate_dims(x.container, [2])[0]

    y = x.container.addVariable(
        _get_random_name("sos1"), type="sos1", domain=[*domain, last_dim]
    )
    eq = x.container.addEquation(
        _get_random_name("eq"),
        domain=domain,
    )

    activation_domain = [*domain, "0"]
    sos1_var_domain = [*domain, "1"]

    eq[...] = y[activation_domain] == x + y[sos1_var_domain]

    if return_slack_var:
        return y[activation_domain], y[sos1_var_domain], [eq]

    return y[activation_domain], [eq]


def relu_with_binary_var(
    x: (
        Parameter
        | Variable
        | implicits.ImplicitParameter
        | implicits.ImplicitVariable
        | Expression
        | Operation
    ),
    default_lb: float = -(10**6),
    default_ub: float = 10**6,
    return_binary_var: bool = False,
):
    """
    Implements the ReLU activation function using binary variables. The ReLU
    function is defined as ReLU(x) = max(x, 0). This implementation **generates**
    one binary variable, one positive variable and three equations. The binary
    variable is necessary to represent the mathematical relationship, while the
    positive variable serves as the activation variable. Both the binary and
    positive variables share the same domain as the input.

    The formulation of this function requires having lower and upper bounds
    for the input ``x``. This function utilizes the bounds from the variables
    if provided. If not, it defaults to the bounds defined by ``default_lb``
    and ``default_ub``. Providing tighter and **correct** bounds can enhance
    the quality of linear relaxations.

    Returns the activation variable and the equation list if ``return_binary_var`` is False,
    otherwise returns activation, binary variable and equation list in order.

    Adapted from `OMLT <https://github.com/cog-imperial/OMLT/blob/e60563859a66ac5dd3348bf1763de57eec95171e/src/omlt/neuralnet/activations/relu.py>`_

    Parameters
    ----------
    x : Parameter | Variable | implicits.ImplicitParameter | implicits.ImplicitVariable | Expression | Operation
    default_ub : float
    default_lb : float
    return_binary_var: bool

    Returns
    -------
    tuple[Variable, list[Equation]] | tuple[Variable, Variable, list[Equation]]

    Examples
    --------
    >>> from gamspy import Container, Variable, Set
    >>> from gamspy.math.activation import relu_with_binary_var
    >>> m = Container()
    >>> i = Set(m, "i", records=range(3))
    >>> x = Variable(m, "x", domain=[i])
    >>> y, eqs = relu_with_binary_var(x)
    >>> y.type
    'positive'
    >>> len(eqs)
    3
    >>> y, b, eqs = relu_with_binary_var(x, return_binary_var=True)
    >>> b.type
    'binary'
    >>> y.domain # i many activation variables
    [Set(name='i', domain=['*'])]
    >>> b.domain # i many binary variables
    [Set(name='i', domain=['*'])]

    """
    assert x.container
    domain = x.domain
    sigma = x.container.addVariable(
        _get_random_name("bin"),
        type="binary",
        domain=domain,
    )

    y = x.container.addVariable(
        _get_random_name("y"),
        type="positive",
        domain=domain,
    )

    eq = [
        x.container.addEquation(
            _get_random_name("eq"),
            domain=domain,
        )
        for _ in range(3)
    ]

    # y >= 0 implied by positive variable
    eq[0][...] = y >= x

    if isinstance(x, Variable):
        eq[1][...] = y <= x - (1 - sigma) * _get_lb(x, default_lb)
        eq[2][...] = y <= sigma * _get_ub(x, default_ub)
        y.lo[...] = gamspy.math.Max(0, x.lo[...])
        y.up[...] = gamspy.math.Max(0, x.up[...])
    else:
        eq[1][...] = y <= x - (1 - sigma) * default_lb
        eq[2][...] = y <= sigma * default_ub
        y.lo[...] = 0

    if return_binary_var:
        return y, sigma, eq
    else:
        return y, eq


def relu_with_complementarity_var(
    x: (
        Parameter
        | Variable
        | implicits.ImplicitParameter
        | implicits.ImplicitVariable
        | Expression
        | Operation
    ),
):
    """
    Implements the ReLU activation function using complementarity conditions.
    The ReLU function is defined as ReLU(x) = max(x, 0). This implementation
    **generates** one positive variable, which serves as the activation variable
    and two equations. The activation variable shares the same domain as the
    input. Unlike ``relu_with_binary_var``, this function does not require lower
    and upper bounds for the formulation.

    Returns the activation variable and the equation list.

    Adapted from `OMLT <https://github.com/cog-imperial/OMLT/blob/e60563859a66ac5dd3348bf1763de57eec95171e/src/omlt/neuralnet/activations/relu.py>`_

    Parameters
    ----------
    x : Parameter | Variable | implicits.ImplicitParameter | implicits.ImplicitVariable | Expression | Operation

    Returns
    -------
    tuple[Variable, list[Equation]]

    Examples
    --------
    >>> from gamspy import Container, Variable, Set
    >>> from gamspy.math.activation import relu_with_complementarity_var
    >>> m = Container()
    >>> i = Set(m, "i", records=range(3))
    >>> x = Variable(m, "x", domain=[i])
    >>> y, eqs = relu_with_complementarity_var(x)
    >>> y.type
    'positive'
    >>> len(eqs)
    2

    """
    assert x.container
    domain = x.domain

    y = x.container.addVariable(
        _get_random_name("y"),
        type="positive",
        domain=domain,
    )

    eq = [
        x.container.addEquation(
            _get_random_name("eq"),
            domain=domain,
        )
        for _ in range(2)
    ]

    # y >= 0 implied by positive variable
    eq[0][...] = y * (y - x) == 0
    eq[1][...] = y - x >= 0

    if isinstance(x, Variable):
        y.lo[...] = gamspy.math.Max(0, x.lo[...])
        y.up[...] = gamspy.math.Max(0, x.up[...])
    else:
        y.lo[...] = 0

    return y, eq


def log_softmax(x: Variable, dim: int = -1, skip_intrinsic: bool = False):
    """
    Implements the log_softmax activation function. This function strictly
    requires a GAMSPy Variable, `y = log_softmax(x)`. The ``dim`` parameter
    specifies the **index** of the softmax dimension. If not provided, it
    calculates log_softmax for the last dimension. This function is preferred
    over the :meth:`softmax <gamspy.math.softmax>` function because, when
    the softmax dimension has 20 or fewer elements, it uses the
    :meth:`lse_max <gamspy.math.lse_max>` (log-sum-exp) intrinsic function for
    improved numerical stability which usually leads to faster solve times.
    Some solvers do not support :meth:`lse_max <gamspy.math.lse_max>`, in that
    case you can set ``skip_intrinsic`` parameter to True to not use intrinsic
    functions even when possible.

    To learn more about `Log-Sum-Exp trick <https://gregorygundersen.com/blog/2020/02/09/log-sum-exp/>`_ .

    This function is usually combined with Negative Log Likelihood loss for
    classification problems.

    Returns the activation variable and the equation list.

    Parameters
    ----------
    x : Variable
    dim : int
    skip_intrinsic: bool

    Returns
    -------
    Variable, list[Equation]

    Examples
    --------
    >>> from gamspy import Container, Variable
    >>> from gamspy.math import dim
    >>> from gamspy.math.activation import log_softmax
    >>> m = Container()
    >>> x = Variable(m, "x", domain=dim([500, 10]))
    >>> y, eqs1 = log_softmax(x) # uses LSE because 10 <= 20
    >>> y.domain
    [Set(name='DenseDim500_1', domain=['*']), Set(name='DenseDim10_1', domain=['*'])]
    >>> y2, eqs2 = log_softmax(x, dim=0) # cannot use LSE because 500 > 20
    >>> y3, eqs3 = log_softmax(x, skip_intrinsic=True) # don't use LSE because of skip_intrinsic

    """
    if not isinstance(x, Variable):
        raise ValidationError("log_softmax expects a variable")

    if dim < 0:
        dim = len(x.domain) + dim

    sum_domain = next_alias(x.domain[dim])

    y = x.container.addVariable(
        _get_random_name("y"),
        domain=x.domain,
    )

    eq = x.container.addEquation(
        _get_random_name("eq"),
        domain=x.domain,
    )

    if not skip_intrinsic and len(sum_domain) != 0 and len(sum_domain) <= 20:
        # Use built-in LSE if possible
        scalars = [rec for rec in sum_domain.records["uni"]]
        variables = []
        for scalar in scalars:
            expr_domain = [*x.domain[:dim], scalar, *x.domain[dim + 1 :]]
            variables.append(x[expr_domain])

        log_sum_exp = gamspy.math.lse_max(*variables)
        eq[...] = y[...] == x - log_sum_exp
    else:
        expr_domain = [
            d if i != dim else sum_domain for (i, d) in enumerate(x.domain)
        ]
        sum_expr = gamspy.Sum(sum_domain, gamspy.math.exp(x[expr_domain]))
        eq[...] = y[...] == x - gamspy.math.log(sum_expr)

    return y, [eq]


def softmax(x: Variable, dim: int = -1):
    """
    Implements the softmax activation function. This function strictly requires
    a GAMSPy Variable, `y = softmax(x)`. The ``dim`` parameter specifies the
    index of the softmax dimension. If not provided, the softmax is calculated
    for the last dimension. This function is implemented for completeness;
    however, in many cases, you can use :meth:`log_softmax <gamspy.math.log_softmax>`
    for better numerical stability.

    Use :meth:`log_softmax <gamspy.math.log_softmax>` if you need to take the
    logarithm of the softmax function.

    Returns the activation variable and the equation list.

    Parameters
    ----------
    x : Variable
    dim : int

    Returns
    -------
    tuple[Variable, list[Equation]]

    Examples
    --------
    >>> from gamspy import Container, Variable
    >>> from gamspy.math import dim
    >>> from gamspy.math.activation import softmax
    >>> m = Container()
    >>> x = Variable(m, "x", domain=dim([500, 10]))
    >>> y, eqs = softmax(x)
    >>> y.domain
    [Set(name='DenseDim500_1', domain=['*']), Set(name='DenseDim10_1', domain=['*'])]

    """
    if not isinstance(x, Variable):
        raise ValidationError("softmax expects a variable")

    if dim < 0:
        dim = len(x.domain) + dim

    sum_domain = next_alias(x.domain[dim])
    expr_domain = [
        d if i != dim else sum_domain for (i, d) in enumerate(x.domain)
    ]

    y = x.container.addVariable(
        _get_random_name("y"),
        domain=x.domain,
    )

    eq = x.container.addEquation(
        _get_random_name("eq"),
        domain=x.domain,
    )

    sum_expr = gamspy.Sum(sum_domain, gamspy.math.exp(x[expr_domain]))
    eq[...] = y[...] == gamspy.math.exp(x) / sum_expr

    return y, [eq]



from __future__ import annotations

from gams.transfer import SpecialValues

import gamspy.formulations as formulations
import gamspy.math as math
import gamspy.utils as utils
from gamspy._algebra import (
    Card,
    Domain,
    Number,
    Ord,
    Product,
    Sand,
    Smax,
    Smin,
    Sor,
    Sum,
)
from gamspy._algebra.expression import Expression
from gamspy._backend.engine import EngineClient
from gamspy._backend.neos import NeosClient
from gamspy._config import _set_default_options, get_option, set_options
from gamspy._container import Container
from gamspy._model import (
    FileFormat,
    Model,
    ModelStatus,
    Problem,
    Sense,
    SolveStatus,
)
from gamspy._options import (
    ConvertOptions,
    FreezeOptions,
    ModelInstanceOptions,
    Options,
)
from gamspy._serialization import deserialize, serialize
from gamspy._symbols import (
    Alias,
    Equation,
    EquationType,
    Parameter,
    Set,
    UniverseAlias,
    Variable,
    VariableType,
)

from .version import __version__

_ctx_managers: dict[tuple[int, int], Container] = dict()
_set_default_options()

__all__ = [
    "Container",
    "Set",
    "Alias",
    "UniverseAlias",
    "Parameter",
    "Variable",
    "Equation",
    "Model",
    "Problem",
    "Sense",
    "VariableType",
    "EquationType",
    "ModelStatus",
    "SolveStatus",
    "Sum",
    "Product",
    "Smax",
    "Smin",
    "Sand",
    "Sor",
    "Domain",
    "Number",
    "Ord",
    "Card",
    "Options",
    "FreezeOptions",
    "ModelInstanceOptions",
    "ConvertOptions",
    "FileFormat",
    "Expression",
    "EngineClient",
    "NeosClient",
    "math",
    "formulations",
    "utils",
    "SpecialValues",
    "__version__",
    "get_option",
    "set_options",
    "serialize",
    "deserialize",
]
